require("common")

relay = {
    -- See also https://ivideon.atlassian.net/wiki/spaces/IE/pages/1399488543/Ivideon+Server+Relay+Outputs+DRAFT?focusedCommentId=1599996122#comment-1599996122
    -- We use ASCII-code to represent one particular state, because it's faster to work with a string by iterating over its bytes
    PAUSE =  string.byte("."),
    INACIVE = string.byte("0"),
    ACTIVE = string.byte("1")
}

function relay.trigger_output(cmd, executor, ...)
    local delay = math.tointeger(cmd.args.delay or 0)
    local port = cmd.args.port
    if not delay or type(port) ~= "string" then
        return {status = 400, reason = reason or "Bad params"}
    end

    local states_seq = cmd.args.state
    for i = 1, #states_seq do
        local state = states_seq:byte(i)
        if state == relay.INACIVE or state == relay.ACTIVE then
            local ok, code, reason = executor(state, port, ...)
            if ok then
                info_log("State was set to " .. string.char(state))
            else
                err_log("Failed to set state to " .. string.char(state))
                return {status = code or 500, reason = reason or "Internal error"}
            end
        elseif state == relay.PAUSE then
            iv.sleep(delay)
        else
            warn_log("Unexpected state " .. string.char(state))
        end
    end

    return {status = 200, reason = "OK"}
end