require("cloud_config_template")

for _,v in ipairs({"streams/2/resolution", "streams/2/bitrate", "streams/2/fps", "streams/2/i_frame_interval", "streams/2/variable_bitrate"}) do settings:remove_base(v) end

DH_MAX_POS = 8192

settings:add({
    { name = "microphone/denoise", meta = {type = "bool"} },
    { name = "motion_detector/0/sensitivity", meta = {type = "int", range = {0, 100, 5} } },
    { name = "motion_detector/0/threshold", meta = {type = "int", range = {0, 100, 1} } },
    { name = "motion_detector/0/threshold", meta = {type = "int", range = {0, 100, 20} } },
    { name = "motion_detector/0/zones", meta = {type = "grid", range = {22, 18} } },
    { name = "osd/datetime/enabled", meta = {type = "bool"} },
    { name = "osd/datetime/display_week", meta = {type = "bool"} },
    { name = "osd/datetime/24h", meta = {type = "bool"} },
    { name = "osd/datetime/format", meta = {type = "choice"} },
    { name = "osd/datetime/pos", meta = {type = "position", range = {DH_MAX_POS, DH_MAX_POS} } },
    { name = "osd/name/enabled", meta = {type = "bool"} },
    { name = "osd/name/text", meta = {type = "string", range = {0, 32}} },
    { name = "osd/name/pos", meta = {type = "position", range = {DH_MAX_POS, DH_MAX_POS} } },

    { name = "image/wdr_level", meta = {type = "int", range = {1, 100} } }
})

last_wdr_level = nil -- special hack to safe last known wdr_level value

settings.shutter_times = { -- spec v.2.82, p. 5.2.1, ms
    ["1/3"] = 333.33, ["1/4"] = 250, ["1/5"] = 200, ["1/6"] = 166.66, ["1/8"] = 125, ["1/10"] = 100, ["1/12"] = 83.33, ["1/15"] = 66.66,
    ["1/25"] = 40, ["1/30"] = 33.33, ["1/50"] = 20, ["1/60"] = 16.66, ["1/100"] = 10, ["1/120"] = 8.33, ["1/250"] = 4, ["1/500"] = 2,
    ["1/1000"] = 1, ["1/2000"] = 0.5, ["1/4000"] = 0.25, ["1/10000"] = 0.1
}

settings.timezones = { -- spec v1.67
    ["0"] = "+00:00", ["1"] = "+01:00", ["2"] = "+02:00", ["3"] = "+03:00", ["4"] = "+03:30", ["5"] = "+04:00", ["6"] = "+04:30", ["7"] = "+05:00", ["8"] = "+05:30",
    ["9"] = "+05:45", ["10"] = "+06:00", ["11"] = "+06:30", ["12"] = "+07:00" , ["13"] = "+08:00", ["14"] = "+09:00", ["15"] = "+09:30", ["16"] = "+10:00",
    ["17"] = "+11:00", ["18"] = "+12:00", ["19"] = "+13:00", ["20"] = "-01:00", ["21"] = "-02:00", ["22"] = "-03:00", ["23"] = "-03:30", ["24"] = "-04:00",
    ["25"] = "-05:00", ["26"] = "-06:00", ["27"] = "-07:00", ["28"] = "-08:00", ["29"] = "-09:00", ["30"] = "-10:00", ["31"] = "-11:00", ["32"] = "-12:00"
}

settings.resolutions = { -- all supported resolutions according to spec v1.67
    D1 = "704x576", HD1 = "352x576", BCIF = "704x288", CIF = "352x288", QCIF = "176x144", VGA = "640x480", QVGA = "320x240",
    SVCD = "480x480", QQVGA = "160x128", SVGA = "800x592", XVGA = "1024x768", WXGA = "1280x800", SXGA = "1280x1024", WSXGA = "1600x1024",
    UXGA = "1600x1200", WUXGA = "1920x1200", ND1 = "240x192", ["720P"] = "1280x720", ["1080P"] = "1920x1080", ["1080N"] = "960x1080"
}

settings.traits = {
    ["streams/0/resolution"] = {
        depends = {"streams/0/fps", "streams/0/i_frame_interval", "streams/0/bitrate", "streams/1/fps", "streams/1/i_frame_interval", "streams/1/resolution"},
        restart_requered = true
    },
    ["streams/1/resolution"] = {
        depends = {"streams/1/bitrate", "streams/1/fps", "streams/1/i_frame_interval"},
        restart_requered = true
    },
    ["streams/0/fps"] = {
        depends = {"streams/0/i_frame_interval", "streams/0/bitrate", "streams/1/fps", "streams/1/i_frame_interval"}
    },
    ["streams/1/fps"] = {
        depends = {"streams/1/bitrate", "streams/1/i_frame_interval"}
    },
    ["streams/0/codec"] = {
        depends = {"streams/0/bitrate", "streams/0/fps", "streams/0/i_frame_interval"},
        restart_requered = true
    },
    ["streams/1/codec"] = {
        depends = {"streams/1/bitrate", "streams/1/fps", "streams/1/i_frame_interval"},
        restart_requered = true
    },
    ["microphone/enabled"] = {
        restart_requered = true
    },
    ["general/timezone"] = {
        restart_requered = true
    },
    ["osd/datetime/enabled"] = {
        restart_requered = true
    },
    ["osd/name/enabled"] = {
        restart_requered = true
    },
    ["image/rotate"] = {
        restart_requered = true
    },
    ["image/wdr"] = {
        depends = {"image/wdr_level"}
    }
}

-- Cover defines the regions which are not visible
-- According to Dahua API docs, the cover zone coordinates values (both width and height) are set in range [0 : 8191].
settings.cover = {
    ZONE_BASE_WIDTH = DH_MAX_POS,
    ZONE_BASE_HEIGHT = DH_MAX_POS
}

---------------------------------------
-- Common functions
---------------------------------------

-- Tricky hack wrapper: filter "Error" responses for Dahua APIs because failture reply contains code 200 and simple text message to indicate fail:
if type(http_request.call_orig) ~= "function" then
    http_request.call_orig = http_request.__call
    http_request.__call = function(...)
        local response, code = http_request.call_orig(...)
        if type(response) == "string" and response:find("Error") then return nil, (code == http_request.OK_CODE and http_request.BAD_REQUEST_CODE or code) end
        return response, code
    end
end

function string:get_ini_value_ex(key, pos, key_escaped)
    if not key_escaped then
        -- Prepare key as regexp, replace magic characters:
        key = key:gsub("[%.%[%]]", function(match) return "%" .. match end)
    end

    -- Skip all space chars after the '=' and capture all the next characters except new line (i.e until new line or end)
    local idx_s, idx_e, val = self:find(key .. "=%s*([^\r\n]+)", pos, false)
    return val, idx_s, idx_e
end

function string:get_ini_value(...)
    local val = self:get_ini_value_ex(...)
    return val
end

function set_cam_param(url, timeout)
    local nil_parm1 = nil
    local nil_parm2 = nil
    local res = http_request(url,nil_part1,nil_parm2,timeout)
    if res and (res:find("OK") ~= nil) then
        return true
    end
    return false
end

function get_cam_property(request_action, response_key)
    local result, status = http_request("/cgi-bin/magicBox.cgi?action=" .. request_action)
    if type(result) == "string" and result:find(response_key) then
        return result:get_ini_value(response_key), status
    end

    return nil, status
end

---------------------------------------
-- Setup functions
---------------------------------------

setup = {
    channel = 0
}

function setup:ch()
    return tostring(self.channel)
end

function setup:set_camera(camera)
    local cam_info = vs.get_camera_info( tonumber(camera) )

    if log_config.tag == tostring(camera) and type(settings["general/online_mode"]) == "table" then return end -- already set
    log_config.tag = tostring(camera)
    if type(cam_info) ~= "table" or string.is_empty(cam_info.urlHigh) then return end

    local channel = cam_info.urlHigh:match("channel=(%d+)&")
    if channel then setup.channel = tonumber(channel) - 1 end
    settings:load_from_storage( tonumber(camera) )
end

function setup:get_online_mode_state()
    local recordConf = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=RecordMode[" .. self:ch() .."]")
    if not recordConf then return end

    local mode = recordConf:get_ini_value("table.RecordMode[" .. self:ch() .."].Mode")
    if mode == "2" then
        settings["general/online_mode"].value = false
    else --if mode == "0" then
        settings["general/online_mode"].value = true
    end
end

function setup:set_online_mode_state()
    local recMode = "2"

    if settings["general/online_mode"].value then
        recMode = "0"
    end

    return set_cam_param("/cgi-bin/configManager.cgi?action=setConfig&RecordMode[" .. self:ch() .."].Mode="..recMode)
end

function setup:set_mute_server_state(newState) -- it's NOT online_mode
    local curState = "on"
    if settings["general/online_mode"].value then curState = "off" end
    if curState == newState then return end

    settings["general/online_mode"].value = (newState ~= "on")
    assert (setup:set_online_mode_state() == true)
end

function setup:get_timezone()
    local ntp_conf = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=NTP")
    if not ntp_conf then return end

    local tz_num = ntp_conf:get_ini_value("table.NTP.TimeZone")
    settings["general/timezone"].value = settings.timezones[tostring(tz_num)]

    if settings["general/timezone"].meta.range then return end
    local r = {}
    for _, tz in pairs(settings.timezones) do r[#r+1] = tz end
    table.sort(r)
    settings["general/timezone"].meta.range = r
end

function setup:set_timezone()
    local set_tz_number
    for tz_num, tz in pairs(settings.timezones) do
        if tz == settings["general/timezone"].value then
            set_tz_number = tz_num
            break
        end
    end

    return set_cam_param("/cgi-bin/configManager.cgi?action=setConfig&NTP.TimeZone=" .. set_tz_number)
end

function setup:get_mic_state()
    if not check_mic_available() then return end
    local audioInputConf = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=Encode[" .. self:ch() .."].MainFormat[0].AudioEnable")
    if not audioInputConf then return end

    local audioEnabled = audioInputConf:get_ini_value("table.Encode[" .. self:ch() .."].MainFormat[0].AudioEnable")
    settings["microphone/enabled"].value = (audioEnabled == "true")
end

function setup:set_mic_state()
    if not check_mic_available() then return false end
    local enabled = tostring(settings["microphone/enabled"].value)
    local params=""

    for i = 0,3 do params = params .. string.format("&Encode[" .. self:ch() .."].MainFormat[%d].AudioEnable=%s", i, enabled) end
    for i = 0,2 do params = params .. string.format("&Encode[" .. self:ch() .."].ExtraFormat[%d].AudioEnable=%s", i, enabled) end

    return set_cam_param("/cgi-bin/configManager.cgi?action=setConfig"..params)
end

function setup:get_ir_state()
    local videoInConf = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=VideoInOptions[" .. self:ch() .."].DayNightColor")
    if not videoInConf then
        settings:remove("night_vision/mode")
        return
    end

    local value = videoInConf:get_ini_value("table.VideoInOptions[" .. self:ch() .. "].DayNightColor")
    if not value then
        settings:remove("night_vision/mode")
    elseif value == "0" then
        settings["night_vision/mode"].value = "off"
    elseif value == "2" then
        settings["night_vision/mode"].value = "on"
    else --if value == "1" then
        settings["night_vision/mode"].value = "auto"
    end
end

function setup:set_ir_state()
    local mode = "1"

    if settings["night_vision/mode"].value == "on" then mode = "2"
    elseif settings["night_vision/mode"].value == "off" then mode = "0"
    else mode = "1"
    end

    return set_cam_param("/cgi-bin/configManager.cgi?action=setConfig&VideoInOptions[" .. self:ch() .. "].DayNightColor="..mode)
end

function setup:get_sound_detector(force)
    if not force and type(settings["sound_detection/enabled"]) ~= "table" then
        return false
    end
    local audiDetectConf = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=AudioDetect[" .. self:ch() .."]")
    if not check_mic_available() or not audiDetectConf then
        settings:remove("sound_detection/enabled")
        settings:remove("sound_detection/sensitivity")
        return false
    end

    local prefix = "table.AudioDetect[" .. self:ch() .."]."

    local enabled = audiDetectConf:get_ini_value(prefix .. "Enable") or "false"
    local anomaly_enabled = audiDetectConf:get_ini_value(prefix .. "AnomalyDetect") or "false"
    local mutation_enabled = audiDetectConf:get_ini_value(prefix .. "MutationDetect") or "false"
    local sens = audiDetectConf:get_ini_value(prefix .. "MinVolume") -- 1..100

    if type(settings["sound_detection/enabled"]) ==  "table" then
        settings["sound_detection/enabled"].value = ( (enabled == "true") or (anomaly_enabled == "true") or (mutation_enabled == "true") )
    end
    if type(settings["sound_detection/sensitivity"]) ==  "table" then
        settings["sound_detection/sensitivity"].value = math.tointeger(101 - tonumber(sens or 1))
        settings["sound_detection/sensitivity"].meta.range = {1, 100, 3}
    end
    return true
end

function setup:set_sound_detector()
    if not check_mic_available() then return false end
    local relUrl = "/cgi-bin/configManager.cgi?action=setConfig"
    local enabledStr = tostring(settings["sound_detection/enabled"].value)
    local prefix = "&AudioDetect[" .. self:ch() .."]."
    relUrl = relUrl .. prefix .. "Enable=" .. enabledStr .. prefix .. "AnomalyDetect=" .. enabledStr .. prefix .. "MutationDetect=" .. enabledStr

    local sens = settings["sound_detection/sensitivity"].value
    if sens < 1 then sens = 1 elseif sens > 100 then sens = 100 end -- range check
    local threshold = 101 - sens -- 1 .. 100
    if threshold < 1 then threshold = 1 end
    relUrl = string.format("%s%sMinVolume=%d%sMutationThreold=%d%sAnomalySensitive=%d", relUrl, prefix, threshold, prefix, threshold, prefix, sens)

    return set_cam_param(relUrl)
end

function setup:get_mic_denoise()
    if not settings["microphone/denoise"] then return end
    local denoise_conf = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=AudioInDenoise[" .. self:ch() .."]")
    if not denoise_conf then
        settings:remove("microphone/denoise")
        return
    end

    local denoise_enabled = denoise_conf:get_ini_value("table.AudioInDenoise[" .. self:ch() .."].enable")
    settings["microphone/denoise"].value = (denoise_enabled == "true")
end

function setup:set_mic_denoise()
    if not settings["microphone/denoise"] then return false end
    local relUrl = "/cgi-bin/configManager.cgi?action=setConfig&AudioInDenoise[" .. self:ch() .."].enable=" .. tostring(settings["microphone/denoise"].value)

    return set_cam_param(relUrl)
end

md.raw_width, md.raw_height = 22, 18 -- TODO get from config!
md.ivideon_height, md.ivideon_width = 8, 12
md.width_k, md.height_k = md.raw_width / md.ivideon_width, md.raw_height / md.ivideon_height
md.use_windows, md.use_version = false, true

function md:dh_to_bingrid(dahua_ini_str)
    local result = ""
    local zero_fill = "0"
    for y = 1, md.raw_height do
        local line_value = 0
        if md.use_windows then line_value = dahua_ini_str:get_ini_value("table.MotionDetect[" .. setup:ch() .."].MotionDetectWindow[0].Region[" .. tostring(y - 1) .. "]")
        else line_value = dahua_ini_str:get_ini_value("table.MotionDetect[" .. setup:ch() .."].Region[" .. tostring(y - 1) .. "]") end
        if not line_value then break end

        local line_in_bin = self:hex2bin( string.format("%x", tonumber(line_value)), true )
        if #line_in_bin > md.raw_width then line_in_bin = line_in_bin:sub(0, md.raw_width)
        else line_in_bin = zero_fill:rep(md.raw_width - #line_in_bin) .. line_in_bin end
        if not md.use_windows then line_in_bin = line_in_bin:reverse() end

        result = result .. line_in_bin
    end
    return result
end

function md:bingrid_to_dh(bingrid, start_y, end_y)
    local response = ""
    -- For API v.3 and higher - use reverse zone (according to doctumentation v2.11):
    local prefix = "&MotionDetect[" .. setup:ch() .."]."
    if md.use_windows and md.use_version then response = response .. prefix .. "DetectVersion=V3.0" end

    for y = start_y or 0, end_y or (md.raw_height-1) do
        local line = bingrid:sub(1 + (y * md.raw_width), (y + 1) * md.raw_width)
        if not md.use_windows then line = line:reverse() end
        line = tonumber("0x" .. self:bin2hex(line))

        if type(line) == "number" then
            if md.use_windows then response = response .. string.format(prefix .. "MotionDetectWindow[0].Region[%d]=%d", y, line)
            else response = response .. string.format(prefix .. "Region[%d]=%d", y, line) end
        end
    end
    return response
end

function setup:get_motion_detector()
    local mdConf = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=MotionDetect")
    local prefix = "table.MotionDetect[" .. self:ch() .."]."
    if type(mdConf) ~= "string" or not settings["motion_detector/enabled"] or not mdConf:get_ini_value(prefix .. "Enable") then
        settings:remove("motion_detector/enabled")
        settings:remove("motion_detector/0/sensitivity")
        settings:remove("motion_detector/0/threshold")
        settings:remove("motion_detector/0/zones")
        return false
    end

    settings["motion_detector/enabled"].value = (mdConf:get_ini_value(prefix .. "Enable") == "true")

    if settings["motion_detector/0/sensitivity"] then
        local value = mdConf:get_ini_value(prefix .. "MotionDetectWindow[0].Sensitive")
        if not string.is_empty(value) then settings["motion_detector/0/sensitivity"].value = tonumber(value)
        else settings["motion_detector/0/sensitivity"].value = (tonumber(mdConf:get_ini_value(prefix .. "Level"))) * 15 end -- 0-6, old method
    end

    if settings["motion_detector/0/threshold"] then
        value = mdConf:get_ini_value(prefix .. "MotionDetectWindow[0].Threshold")
        if not string.is_empty(value) then settings["motion_detector/0/threshold"].value = tonumber(value)
        else settings:remove("motion_detector/0/threshold") end
    end

    if settings["motion_detector/0/zones"] then
        if mdConf:get_ini_value(prefix .. "MotionDetectWindow[0].Region[0]") then md.use_windows = true end
        local bingrid = md:dh_to_bingrid(mdConf)
        if not string.is_empty(bingrid) then settings["motion_detector/0/zones"].value = bingrid
        else settings:remove("motion_detector/0/zones") end
    end
    return true
end

function setup:set_motion_detector()
    local relUrl = "/cgi-bin/configManager.cgi?action=setConfig"
    local prefix = "&MotionDetect[" .. self:ch() .."]."
    if settings["motion_detector/enabled"] then relUrl = relUrl .. prefix .. "Enable=" .. tostring(settings["motion_detector/enabled"].value) end

    if settings["motion_detector/0/sensitivity"] then
        if md.use_windows then relUrl = relUrl .. prefix .. "MotionDetectWindow[0].Sensitive=" .. tostring(settings["motion_detector/0/sensitivity"].value)
        else relUrl = relUrl .. prefix .. "Level=" .. tostring(settings["motion_detector/0/sensitivity"].value / 15) end
    end

    if settings["motion_detector/0/threshold"] then relUrl = relUrl .. prefix .. "MotionDetectWindow[0].Threshold=" .. tostring(settings["motion_detector/0/threshold"].value) end

    return set_cam_param(relUrl)
end

function setup:set_motion_detector_zone_impl()
    if not settings["motion_detector/0/zones"] then return false end

    local middle_y = round(md.raw_height / 2)
    local url1 = "/cgi-bin/configManager.cgi?action=setConfig" .. md:bingrid_to_dh(settings["motion_detector/0/zones"].value, nil, middle_y)
    local url2 = "/cgi-bin/configManager.cgi?action=setConfig" .. md:bingrid_to_dh(settings["motion_detector/0/zones"].value, middle_y, nil)

    return set_cam_param(url1) and set_cam_param(url2)
end

function setup:set_motion_detector_zone()
    local result = setup:set_motion_detector_zone_impl()
    -- Work-around for cameras wich doesn't recognize "DetectVersion=" statement
    if not result and md.use_windows then
        md.use_version = not md.use_version
        result = setup:set_motion_detector_zone_impl()
    end

    return result
end

function setup:get_pir_settings()
    local pirConf = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=Alarm[1].Enable")
    if not pirConf then
        settings:remove("pir_detection/enabled")
        return
    end

    settings["pir_detection/enabled"].value = (pirConf:get_ini_value("table.Alarm[1].Enable") == "true")
end

function setup:set_pir_settings()
    if not settings["pir_detection/enabled"] then return false end
    local enabled = settings["pir_detection/enabled"].value
    return set_cam_param("/cgi-bin/configManager.cgi?action=setConfig&Alarm[1].SenseMethod=PassiveInfrared&Alarm[1].Enable=" .. tostring(enabled))
end

function setup:set_mic_sensitivity()
    if not check_mic_available() then return false end
    local result = false
    for i = 0, 3 do
        local url = "/cgi-bin/configManager.cgi?action=setConfig&AudioInputVolume[" .. tostring(i)  .. "]=" .. tostring(settings["microphone/sensitivity"].value)
        if not set_cam_param(url) then break end
        result = true
    end
    return result
end

function setup:get_mic_sensitivity()
    if not check_mic_available() then return end
    local mdConf = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=AudioInputVolume")
    local value = nil
    if type(mdConf) == "string" then value = mdConf:get_ini_value("table.AudioInputVolume[0]") end

    if not value then
        settings:remove("microphone/sensitivity")
        return
    end
    settings["microphone/sensitivity"].value = tonumber(value)
    settings["microphone/sensitivity"].meta.range = {0, 100}
end

function setup:set_power_freq()
    local flicker_value = "0"
    if settings["general/power_freq"].value == "50hz" then flicker_value = "1"
    elseif settings["general/power_freq"].value == "60hz" then flicker_value = "2" end

    local relUrl = "/cgi-bin/configManager.cgi?action=setConfig&VideoInOptions[" .. self:ch() .."].AntiFlicker=" .. flicker_value ..
    "&VideoInOptions[" .. self:ch() .."].NightOptions.AntiFlicker=" .. flicker_value .. "&VideoInOptions[" .. self:ch() .."].NormalOptions.AntiFlicker=" .. flicker_value
    return set_cam_param(relUrl)
end

function setup:get_power_freq()
    local freq_conf = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=VideoInOptions")
    if type(freq_conf) == "string" then freq_conf = freq_conf:get_ini_value("table.VideoInOptions[" .. self:ch() .."].AntiFlicker") end
    if not freq_conf then
        settings:remove("general/power_freq")
        return
    end

    if freq_conf == "0" then  settings["general/power_freq"].value = "auto"
    elseif freq_conf == "1" then settings["general/power_freq"].value = "50hz"
    else settings["general/power_freq"].value = "60hz" end
    settings["general/power_freq"].meta.range = {"auto", "50hz", "60hz"}
end

function setup:get_streams_settings()
    local stream_tags = {[0] = "MainFormat[0]", [1] = "ExtraFormat[0]"}
    local supported = http_request("/cgi-bin/encode.cgi?action=getConfigCaps&channel=" .. self:ch()) -- FIXME get channelId from video URL
    if not supported then return end
    local capa_prefix = "caps.caps[0]." -- 0 is constant!
    if supported:find(capa_prefix, 1, true) == nil then capa_prefix = "caps[0]." end

    for i = 0, 1 do
        local dahua_tag = "Encode[" .. self:ch() .. "]." .. stream_tags[i]
        local encode_config = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=" .. dahua_tag)
        if not encode_config then return end

        local setting_prefix = "streams/" .. tostring(i) .. "/"
        local vid_tag = "table." .. dahua_tag .. ".Video."
        local aud_tag = "table." .. dahua_tag .. ".Audio."

        settings[setting_prefix .. "bitrate"].value = tonumber( encode_config:get_ini_value(vid_tag .. "BitRate") )
        local bitrate_opts = supported:get_ini_value(capa_prefix .. stream_tags[i] .. ".Video.BitRateOptions")
        if bitrate_opts == nil then
            err_log("Bad bitrate options for " .. tostring(i).." stream: " .. supported)
            return
        end
        local bitrates = bitrate_opts:split(",")
        local bitrate_min, bitrate_max = tonumber(bitrates[1]), tonumber(bitrates[2])
        local bitrate_range = {128, 256, 512, 768, 1024, 1536, 2048, 3072, 4096, 6144, 8192}
        for j = #bitrate_range, 1, -1 do
            if bitrate_range[j] < bitrate_min or bitrate_max < bitrate_range[j] then
                table.remove(bitrate_range, j)
            end
        end
        if not is_in_table(settings[setting_prefix .. "bitrate"].value, bitrate_range) then
            bitrate_range[#bitrate_range+1] = settings[setting_prefix .. "bitrate"].value
        end
        settings[setting_prefix .. "bitrate"].meta.range = bitrate_range

        settings[setting_prefix .. "variable_bitrate"].value = (encode_config:get_ini_value(vid_tag .. "BitRateControl") == "VBR")

        settings[setting_prefix .. "fps"].value = tonumber( encode_config:get_ini_value(vid_tag .. "FPS") )
        local fps_max = tonumber( supported:get_ini_value(capa_prefix .. stream_tags[i] .. ".Video.FPSMax") )
        local fps_range = {1, 2, 3, 5, 7, 10, 12, 15, 20, 25, 30, 40, 45, 50, 60}
        for j = #fps_range, 1, -1 do
            if fps_max < fps_range[j] then table.remove(fps_range, j) end
        end
        if not is_in_table(settings[setting_prefix .. "fps"].value, fps_range) then
            fps_range[#fps_range+1] = settings[setting_prefix .. "fps"].value
        end
        settings[setting_prefix .. "fps"].meta.range = fps_range

        settings[setting_prefix .. "i_frame_interval"].value = tonumber( encode_config:get_ini_value(vid_tag .. "GOP") )
        settings[setting_prefix .. "i_frame_interval"].meta.range = {1, settings[setting_prefix .. "fps"].value * 4}

        settings[setting_prefix .. "resolution"].value = encode_config:get_ini_value(vid_tag .. "Width") .. "x" .. encode_config:get_ini_value(vid_tag .. "Height")
        local res_range = {}
        local supported_res = supported:get_ini_value(capa_prefix .. stream_tags[i] .. ".Video.ResolutionTypes")
        supported_res = supported_res:split(",")
        for j = 1, #supported_res do
            if supported_res[j]:find("x") ~= nil then
                if not is_in_table(supported_res[j], res_range) then res_range[#res_range+1] = supported_res[j] end
            elseif settings.resolutions[ supported_res[j] ] ~= nil then
                local cur_res = settings[setting_prefix .. "resolution"].value:split("x")
                local new_option = settings.resolutions[ supported_res[j] ]:split("x")
                -- if i == 1 then -- work-around only for substream!
                --     new_option[2] = round( ( cur_res[2] * new_option[1] ) / ( cur_res[1] * 8) ) * 8 -- Y transformation in respect to aspect ratio, round to 8
                -- end
                new_option = tostring(new_option[1]) .. "x" .. tostring(new_option[2])
                if not is_in_table(new_option, res_range) then res_range[#res_range+1] = new_option end
            end
        end
        if not is_in_table(settings[setting_prefix .. "resolution"].value, res_range) then
            res_range[#res_range+1] = settings[setting_prefix .. "resolution"].value
        end
        settings[setting_prefix .. "resolution"].meta.range = res_range

        if settings[setting_prefix .. "codec"] then
            settings[setting_prefix .. "codec"].value = encode_config:get_ini_value(vid_tag .. "Compression")

            local codecs_range = {}
            local codecs = supported:get_ini_value(capa_prefix .. stream_tags[i] .. ".Video.CompressionTypes")
            if codecs ~= nil then
                codecs = codecs:split(",")
                for j = 1, #codecs do
                    if not codecs[j]:find("JP") then codecs_range[j] = codecs[j] end
                end
            end
            if #codecs_range <= 1 then settings:remove(setting_prefix .. "codec")
            else settings[setting_prefix .. "codec"].meta.range = codecs_range end
        end

        if settings[setting_prefix .. "acodec"] then
            settings[setting_prefix .. "acodec"].value = encode_config:get_ini_value(aud_tag .. "Compression")

            local codecs_range = {}
            local codecs = supported:get_ini_value(capa_prefix .. stream_tags[i] .. ".Audio.CompressionTypes")
            if codecs ~= nil then
                codecs = codecs:split(",")
                for j = 1, #codecs do codecs_range[j] = codecs[j] end
            end
            if #codecs_range <= 1 then settings:remove(setting_prefix .. "acodec")
            else settings[setting_prefix .. "acodec"].meta.range = codecs_range end
        end
    end
end

function setup:set_stream_settings(streamId)
    -- Setup format: Encode[Channel].MainFormat[RecordType] for primary stream, where
    -- RecordType is: 0 - live/continuous, 1 - motion, 2 - alarm record
    -- Encode[Channel].ExtraFormat[ExtraStream] for secondary stream
    -- We'll set all streams to the same settings for compatibility
    local setting_prefix = "streams/" .. tostring(streamId) .. "/"
    local stream_prefixes = {[0] = "&Encode[" .. self:ch() .. "].MainFormat[%d]", [1] = "&Encode[" .. self:ch() .. "].ExtraFormat[%d]"}
    local resolution = settings[setting_prefix .. "resolution"].value:split("x")

    -- Detect maximum supported FPS:
    local fps = settings[setting_prefix .. "fps"].value
    local auto_gop = 2 -- 2 sec by default; use 3 seconds to allow less bitrate on huge pictures
    if tonumber( resolution[1] ) > 2000 then auto_gop = 3 end

    local i_frame_interval = fps * auto_gop
    if settings[setting_prefix .. "i_frame_interval"].changed then i_frame_interval = settings[setting_prefix .. "i_frame_interval"].value end

    local stream_for_test = string.format(stream_prefixes[streamId], 0) .. ".Video"
    stream_for_test = "&Encode[0" .. stream_for_test:sub(10) -- FIXME XVR / ANALOGUE
    local capabilities = http_request("/cgi-bin/encode.cgi?action=getConfigCaps&channel=" .. self:ch() ..
        stream_for_test .. ".Width=" .. resolution[1] ..
        stream_for_test .. ".Height=" .. resolution[2] ..
        stream_for_test .. ".FPS=" .. tostring(fps) ..
        stream_for_test .. ".GOP=" .. tostring(i_frame_interval))

    if capabilities ~= nil then
        local capa_prefix = "caps.caps"
        if capabilities:find(capa_prefix, 1, true) == nil then capa_prefix = "caps" end

        local max_fps = tonumber( capabilities:get_ini_value(capa_prefix .. stream_for_test:sub(8) .. ".FPSMax") )
        if max_fps ~= nil and max_fps < fps then fps = max_fps end

        local bitrates = capabilities:get_ini_value(capa_prefix .. stream_for_test:sub(8) .. ".BitRateOptions"):split(",")
        local bitrate_min, bitrate_max = tonumber(bitrates[1]), tonumber(bitrates[2])
        if settings[setting_prefix .. "bitrate"].value < bitrate_min then
            settings[setting_prefix .. "bitrate"].value = bitrate_min
        end
        if settings[setting_prefix .. "bitrate"].value > bitrate_max then
            settings[setting_prefix .. "bitrate"].value = bitrate_max
        end
    end

    local bitrate_type = (settings[setting_prefix .. "variable_bitrate"].value) and "VBR" or "CBR"

    local req_skel = "TAG.Video.BitRateControl=" .. bitrate_type ..
                     "TAG.Video.BitRate=" .. tostring(settings[setting_prefix .. "bitrate"].value) ..
                     "TAG.Video.FPS=" .. tostring(fps) .. "TAG.Video.GOP=" .. tostring(i_frame_interval) ..
                     "TAG.Video.Width=" .. resolution[1] ..
                     "TAG.Video.Height=" .. resolution[2] ..
                     ""
    if settings[setting_prefix .. "codec"] then req_skel = req_skel .. "TAG.Video.Compression=" .. settings[setting_prefix .. "codec"].value end
    if settings[setting_prefix .. "acodec"] then req_skel = req_skel .. "TAG.Audio.Compression=" .. settings[setting_prefix .. "acodec"].value end
    iv.log("Set stream " .. tostring(streamId) .. " settings: " .. req_skel)

    -- set success mark only with respect to primary (streaming) config to prevent fake success results
    local success = set_cam_param("/cgi-bin/configManager.cgi?action=setConfig" ..  req_skel:gsub( "TAG", string.format(stream_prefixes[streamId], 0) ))
    if not success and tonumber( resolution[1] ) >= 2304 and tonumber(fps) > 20 then -- tricky hack for FPSMax issue
        settings[setting_prefix .. "fps"].value = 20
        return self:set_stream_settings(streamId)
    end
    local substream_count = (streamId == 0) and 3 or 2
    for j = 1, substream_count do set_cam_param("/cgi-bin/configManager.cgi?action=setConfig" ..  req_skel:gsub( "TAG", string.format(stream_prefixes[streamId], j) )) end
    return success
end

function setup:get_osd_settings()
    local widget = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=VideoWidget[" .. self:ch() .. "]")
    local base_tag = "table.VideoWidget[" .. self:ch() .. "]"
    local name = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=ChannelTitle[" .. self:ch() .. "]") or ""
    local time_format = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=Locales.TimeFormat") or ""

    if not widget or type(widget) ~= "string" then
        settings:remove("osd/datetime/enabled")
        settings:remove("osd/datetime/display_week")
        settings:remove("osd/datetime/24h")
        settings:remove("osd/datetime/format")
        settings:remove("osd/datetime/pos")
        settings:remove("osd/name/enabled")
        settings:remove("osd/name/text")
        settings:remove("osd/name/pos")
        return
    end

    local cs = widget:get_ini_value(base_tag .. ".ChannelTitle.EncodeBlend")
    if settings["osd/name/enabled"] and type(cs) == "string" then
        settings["osd/name/enabled"].value = (cs == "true")
    else
        settings:remove("osd/name/enabled")
        settings:remove("osd/name/text")
        settings:remove("osd/name/pos")
    end
    cs = name:get_ini_value("table.ChannelTitle[" .. self:ch() .. "].Name")
    if settings["osd/name/text"] and type(cs) == "string" then
        settings["osd/name/text"].value = cs
    else
        settings:remove("osd/name/text")
    end
    if settings["osd/name/pos"] then
        settings["osd/name/pos"].value = {
            tonumber( widget:get_ini_value(base_tag .. ".ChannelTitle.Rect[0]") or 0 ),
            DH_MAX_POS - tonumber( widget:get_ini_value(base_tag .. ".ChannelTitle.Rect[1]") or 0 )
        }
    end

    cs = widget:get_ini_value(base_tag .. ".TimeTitle.EncodeBlend")
    if settings["osd/datetime/enabled"] and type(cs) == "string" then
        settings["osd/datetime/enabled"].value = (cs == "true")
    else
        settings:remove("osd/datetime/enabled")
        settings:remove("osd/datetime/display_week")
        settings:remove("osd/datetime/24h")
        settings:remove("osd/datetime/format")
        settings:remove("osd/datetime/pos")
    end
    if settings["osd/datetime/pos"] then
        settings["osd/datetime/pos"].value = {
            tonumber( widget:get_ini_value(base_tag .. ".TimeTitle.Rect[0]") or 0 ),
            DH_MAX_POS - tonumber( widget:get_ini_value(base_tag .. ".TimeTitle.Rect[1]") or 0 )
        }
    end
    cs = widget:get_ini_value(base_tag .. ".TimeTitle.ShowWeek")
    if settings["osd/datetime/display_week"] and type(cs) == "string" then
        settings["osd/datetime/display_week"].value = (cs == "true")
    else
        settings:remove("osd/datetime/display_week")
    end

    time_format = time_format:get_ini_value("table.Locales.TimeFormat")
    if settings["osd/datetime/24h"] and type(time_format) == "string" then
        settings["osd/datetime/24h"].value = (time_format:find("HH") ~= nil)
        settings["osd/datetime/format"].value = time_format:gsub("HH", "hh") -- transform HH to hh
        local r = {"yyyy-MM-dd hh:mm:ss", "yyyy/MM/dd hh:mm:ss", "dd/MM/yyyy hh:mm:ss"}
        if not is_in_table(settings["osd/datetime/format"].value, r) then r[#r + 1] = settings["osd/datetime/format"].value end
        settings["osd/datetime/format"].meta.range = r
    else
        settings:remove("osd/datetime/24h")
        settings:remove("osd/datetime/format")
    end
    -- for i in range (CustomTitle)
end

function setup:set_osd_settings()
    local function position_set(cur_config, base_tag, value)
        if type(cur_config) ~= "string" or type(value) ~= "table" or type(value[2]) ~= "number" then return "" end
        local cc_tag = "table." .. base_tag
        local width, height = tonumber(cur_config:get_ini_value(cc_tag .. "[0]") or 0) - tonumber(cur_config:get_ini_value(cc_tag .. "[2]") or 0),
            tonumber(cur_config:get_ini_value(cc_tag .. "[1]") or 0) - tonumber(cur_config:get_ini_value(cc_tag .. "[3]") or 0)
        if value[1] + width >= DH_MAX_POS then value[1] = DH_MAX_POS - width end
        if value[2] - height <= 0 then value[2] = height end
        return base_tag .. "[0]=" .. tostring(value[1]) .. base_tag .. "[1]=" .. tostring(DH_MAX_POS - value[2]) ..
               base_tag .. "[2]=" .. tostring(value[1] + width) .. base_tag .. "[3]=" .. tostring(DH_MAX_POS - value[2] + height)
    end

    local cur_config = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=VideoWidget[" .. self:ch() .. "]")
    local rel_url = "/cgi-bin/configManager.cgi?action=setConfig"
    local result = true
    if settings["osd/name/text"] then result = set_cam_param(rel_url .. "&ChannelTitle[" .. self:ch() .. "].Name=" .. iv.escape_url(settings["osd/name/text"].value)) and result end

    if settings["osd/datetime/format"] then
        local tf = settings["osd/datetime/format"].value
        if settings["osd/datetime/24h"].value then tf = tf:gsub("hh", "HH") end
        result = set_cam_param(rel_url .. "&Locales.TimeFormat=" .. iv.escape_url(tf)) and result
    end

    local base_tag = "&VideoWidget[" .. self:ch() .. "]"
    if settings["osd/name/enabled"] then rel_url = rel_url .. base_tag .. ".ChannelTitle.EncodeBlend=" .. tostring(settings["osd/name/enabled"].value) end
    if settings["osd/name/pos"] then rel_url = rel_url .. position_set(cur_config, base_tag .. ".ChannelTitle.Rect", settings["osd/name/pos"].value) end

    if settings["osd/datetime/enabled"] then rel_url = rel_url .. base_tag .. ".TimeTitle.EncodeBlend=" .. tostring(settings["osd/datetime/enabled"].value) end
    if settings["osd/datetime/pos"] then rel_url = rel_url .. position_set(cur_config, base_tag .. ".TimeTitle.Rect", settings["osd/datetime/pos"].value) end
    if settings["osd/datetime/display_week"] then rel_url = rel_url .. base_tag .. ".TimeTitle.ShowWeek=" .. tostring(settings["osd/datetime/display_week"].value) end

    return set_cam_param(rel_url) and result
end

function setup:get_color_settings()
    local pathSkel = "VideoColor[" .. self:ch() .. "][0]" -- "normal" part of config; 1 - day, 2 - night
    local colorConf = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=" .. pathSkel)
    if string.is_empty(colorConf) then
        settings:remove("image/brightness")
        settings:remove("image/contrast")
        settings:remove("image/saturation")
        settings:remove("image/hue")
        return
    end
    local tmpVal = nil

    tmpVal = colorConf:get_ini_value("table." .. pathSkel .. ".Brightness")
    if tmpVal then settings["image/brightness"].value = tonumber(tmpVal)
    else settings:remove("image/brightness") end
    tmpVal = colorConf:get_ini_value("table." .. pathSkel .. ".Contrast")
    if tmpVal then settings["image/contrast"].value = tonumber(tmpVal)
    else settings:remove("image/contrast") end
    tmpVal = colorConf:get_ini_value("table." .. pathSkel .. ".Saturation")
    if tmpVal then settings["image/saturation"].value = tonumber(tmpVal)
    else settings:remove("image/saturation") end
    tmpVal = colorConf:get_ini_value("table." .. pathSkel .. ".Hue")
    if tmpVal then settings["image/hue"].value = tonumber(tmpVal)
    else settings:remove("image/hue") end
end

function setup:set_color_settings()
    local relUrl = "/cgi-bin/configManager.cgi?action=setConfig"
    local pathSkel = "&VideoColor[" .. self:ch() .. "]["

    if settings["image/brightness"] then
        for i = 0, 2 do relUrl = relUrl .. pathSkel .. tostring(i) .. "].Brightness=" .. tostring(settings["image/brightness"].value) end
    end
    if settings["image/contrast"] then
        for i = 0, 2 do relUrl = relUrl .. pathSkel .. tostring(i) .. "].Contrast=" .. tostring(settings["image/contrast"].value) end
    end
    if settings["image/saturation"] then
        for i = 0, 2 do relUrl = relUrl .. pathSkel .. tostring(i) .. "].Saturation=" .. tostring(settings["image/saturation"].value) end
    end
    if settings["image/hue"] then
        for i = 0, 2 do relUrl = relUrl .. pathSkel .. tostring(i) .. "].Hue=" .. tostring(settings["image/hue"].value) end
    end
    return set_cam_param(relUrl)
end

function setup:reset_exposure_settings()
    -- internal setting that breaks image brightness on modern firmwares after upgrade
    local relUrl = "/cgi-bin/configManager.cgi?action=restore" -- setConfig for dirt way
    for i = 0, 2 do
        -- relUrl = relUrl .. "&VideoInExposure[" .. self:ch() .. "][" .. tostring(i) .. "].Compensation=50" -- dirt way
        relUrl = relUrl .. "&names[" .. tostring(i) .. "]=VideoInExposure[" .. self:ch() .. "][" .. tostring(i) .. "].Compensation"
    end

    return set_cam_param(relUrl)
end

function setup:get_exposure_settings()
    local pathSkel = "VideoInExposure[" .. self:ch() .. "][0]" -- 0 - default, 1 - night, 2 - normal
    local exposureConf = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=" .. pathSkel)
    if string.is_empty(exposureConf) then
        settings:remove("image/shutter_time")
        settings:remove("image/gain")
        return
    end

    local gain_max = exposureConf:get_ini_value("table." .. pathSkel .. ".GainMax")
    if not string.is_empty(gain_max) then
        settings["image/gain"].value = tonumber(gain_max)
        settings["image/gain"].meta.range = {0, 100} -- GainMin <= Gain <= GainMax
    else settings:remove("image/gain") end

    settings["image/shutter_time"].value = nil
    local st_min = tonumber( exposureConf:get_ini_value("table." .. pathSkel .. ".Value1") or 0.0) -- works in manual or auto
    local st_max = tonumber( exposureConf:get_ini_value("table." .. pathSkel .. ".Value2") or 0.0) -- works in auto
    local min_diff = 1000.0
    for k,v in pairs(settings.shutter_times) do
        local diff = (v < st_max) and st_max - v or v - st_max
        if diff <= min_diff then
            min_diff = diff
            settings["image/shutter_time"].value = k
        end
        if diff == 0.0 then break end
    end
    if not settings["image/shutter_time"].meta.range then
        local st_range = {}
        st_range[#st_range + 1] = "auto" -- mode 0
        st_range[#st_range + 1] = "gain" -- mode 7
        for k,v in pairs(settings.shutter_times) do st_range[#st_range + 1] = k end
        settings["image/shutter_time"].meta.range = st_range
    end

    local expo_type = exposureConf:get_ini_value("table." .. pathSkel .. ".Mode")
    if is_in_table(expo_type, {"4", "6", "8"}) then
        iv.log("Detected manual expo mode " .. tostring(expo_type))
    elseif expo_type == "7" then
        settings["image/shutter_time"].value = "gain"
    else settings["image/shutter_time"].value = "auto" -- force default
    end
end

function setup:set_exposure_settings()
    local relUrl = "/cgi-bin/configManager.cgi?action=setConfig"
    local pathSkel = "&VideoInExposure[" .. self:ch() .. "]["

    if settings["image/gain"] and settings["image/gain"].changed then for i = 0, 2 do
        relUrl = relUrl .. pathSkel .. tostring(i) .. "].Gain=" .. tostring(settings["image/gain"].value)
                .. pathSkel .. tostring(i) .. "].GainMin=0"
                .. pathSkel .. tostring(i) .. "].GainMax=" .. tostring(settings["image/gain"].value)
    end end
    if settings["image/shutter_time"] and settings["image/shutter_time"].changed then
        local new_mode = settings["image/shutter_time"].value
        local new_st = nil
        if new_mode == "auto" then new_mode = "1"
        elseif new_mode == "gain" then new_mode = "7"
        else
            new_st = settings.shutter_times[settings["image/shutter_time"].value]
            new_mode = "4" -- manual range, fixed (6) doesn't work on most devices
        end

        for i = 0, 2 do
            relUrl = relUrl .. pathSkel .. tostring(i) .. "].Mode=" .. new_mode
            if new_st ~= nil then relUrl = relUrl .. pathSkel .. tostring(i) .. "].Value1=" .. tostring(new_st)
                .. pathSkel .. tostring(i) .. "].Value2=" .. tostring(new_st) end
        end
    end

    return set_cam_param(relUrl)
end

function setup:get_wb_settings()
    local pathSkel = "VideoInWhiteBalance[" .. self:ch() .. "][0]" -- 0 - default, 1 - night, 2 - normal
    local wbConf = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=" .. pathSkel)
    if string.is_empty(wbConf) then
        settings:remove("image/wb")
        return
    end

    settings["image/wb"].value = wbConf:get_ini_value("table." .. pathSkel .. ".Mode")
    settings["image/wb"].meta.range = {"Auto", "Natural", "Street Lamp", "Outdoor", "Manual"}
    --local wb_range = {0, 100}
    --settings["image/wb_red"].value = wbConf:get_ini_value("table." .. pathSkel .. ".GainRed")
    --settings["image/wb_red"].meta.range = wb_range
    --settings["image/blue"].value = wbConf:get_ini_value("table." .. pathSkel .. ".GainBlue")
    --settings["image/wb_blue"].meta.range = wb_range
end

function setup:set_wb_settings()
    local relUrl = "/cgi-bin/configManager.cgi?action=setConfig"
    local pathSkel = "&VideoInWhiteBalance[" .. self:ch() .. "]["

    if settings["image/wb"] and settings["image/wb"].changed then
        for i = 0, 2 do relUrl = relUrl .. pathSkel .. tostring(i) .. "].Mode=" .. iv.escape_url(settings["image/wb"].value) end
    end
    if settings["image/wb_red"] and settings["image/wb_red"].changed then
        for i = 0, 2 do relUrl = relUrl .. pathSkel .. tostring(i) .. "].GainRed=" .. tostring(settings["image/wb_red"].value) end
    end
    if settings["image/wb_blue"] and settings["image/wb_blue"].changed then
        for i = 0, 2 do relUrl = relUrl .. pathSkel .. tostring(i) .. "].GainBlue=" .. tostring(settings["image/wb_blue"].value) end
    end

    return set_cam_param(relUrl)
end

function setup:get_videoin_settings()
    local pathSkel = "VideoInOptions[" .. self:ch() .. "]"
    local videoInConf = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=" .. pathSkel)
    if string.is_empty(videoInConf) then
        settings:remove("image/upside_down")
        settings:remove("image/rotate")
        return
    end
    local tmpVal = nil

    tmpVal = videoInConf:get_ini_value("table." .. pathSkel .. ".Mirror")
    if not string.is_empty(tmpVal) then settings["image/upside_down"].value = (tmpVal == "true")
    else settings:remove("image/upside_down") end

    tmpVal = videoInConf:get_ini_value("table." .. pathSkel .. ".Rotate90")
    if tmpVal then settings["image/rotate"].value = (tmpVal == "1")
    else settings:remove("image/rotate") end
end

function setup:set_image_control()
    -- new-gen rotation settings (5.1.3)
    local relUrl = "/cgi-bin/configManager.cgi?action=setConfig"
    local pathSkel = "&VideoImageControl[" .. self:ch() .. "]"

    if settings["image/rotate"] and settings["image/rotate"].changed then
        local rotateMode = settings["image/rotate"].value and "1" or "0"
        relUrl = relUrl .. pathSkel .. ".Rotate90=" .. rotateMode
    end
    if settings["image/upside_down"] and settings["image/upside_down"].changed then
        relUrl = relUrl .. pathSkel .. ".Mirror=" .. tostring(settings["image/upside_down"].value)
    end
    local need_flip = (settings["image/upside_down"] and settings["image/upside_down"].value) and not (settings["image/rotate"] and settings["image/rotate"].value)
    relUrl = relUrl .. pathSkel .. ".Flip=" .. tostring(need_flip)

    return set_cam_param(relUrl)
end

function setup:set_videoin_settings()
    local relUrl = "/cgi-bin/configManager.cgi?action=setConfig"
    local pathSkel = "&VideoInOptions[" .. self:ch() .. "]" -- fixme overloaded in VideoInExplosure (5.2.1) secton too?
    self:set_image_control()

    if settings["image/rotate"] and settings["image/rotate"].changed then
        local rotateMode = "0" -- 1 - 90 degrees, 2 - 270 degrees (mode=2 == mirror + 1)
        if settings["image/rotate"].value then rotateMode = "1" end
        for _, mode in ipairs({"", ".NightOptions", ".NormalOptions"}) do
            relUrl = relUrl .. pathSkel .. mode .. ".Rotate90=" .. rotateMode
        end
    end
    if settings["image/upside_down"] and settings["image/upside_down"].changed then
        for _, mode in ipairs({"", ".NightOptions", ".NormalOptions"}) do
            relUrl = relUrl .. pathSkel .. mode .. ".Mirror=" .. tostring(settings["image/upside_down"].value)
        end
    end
    local need_flip = (settings["image/upside_down"] and settings["image/upside_down"].value) and not (settings["image/rotate"] and settings["image/rotate"].value)
    relUrl = relUrl .. pathSkel .. ".Flip=" .. tostring(need_flip)

    return set_cam_param(relUrl)
end

function setup:get_videoin_backlight()
    local pathSkel = "VideoInBacklight[" .. self:ch() .. "][0]" -- 0 - Normal, 1 - Day, 2 - Night
    local backlightConf = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=" .. pathSkel)
    if string.is_empty(backlightConf) then
        settings:remove("image/wdr")
        settings:remove("image/wdr_level")
        return
    end

    local tmpVal = backlightConf:get_ini_value("table." .. pathSkel .. ".Mode")
    if not string.is_empty(tmpVal) then
        settings["image/wdr"].value = (tmpVal == "WideDynamic")
    else
        settings:remove("image/wdr")
        settings:remove("image/wdr_level")
        return
    end

    tmpVal = backlightConf:get_ini_value("table." .. pathSkel .. ".WideDynamicRange")
    if not string.is_empty(tmpVal) then settings["image/wdr_level"].value = tonumber(tmpVal)
    else settings:remove("image/wdr_level") end
end

function setup:set_videoin_backlight()
    local relUrl = "/cgi-bin/configManager.cgi?action=setConfig"
    local pathSkel = "&VideoInBacklight[" .. self:ch() .. "]["

    if settings["image/wdr"] and settings["image/wdr"].changed then
        local new_mode = (settings["image/wdr"].value) and "WideDynamic" or "Off"
        for i = 0, 2 do relUrl = relUrl .. pathSkel .. tostring(i) .. "].Mode=" .. new_mode end
    end

    if settings["image/wdr_level"] and settings["image/wdr_level"].changed then
        for i = 0, 2 do relUrl = relUrl .. pathSkel .. tostring(i) .. "].WideDynamicRange=" .. tostring(settings["image/wdr_level"].value) end
    end

    return set_cam_param(relUrl)
end

function setup:get_thermo_settings()
    local HITConfig = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=HeatImagingThermometry")
    if string.is_empty(HITConfig) then return end

    return HITConfig:get_ini_value("table.HeatImagingThermometry.ObjectEmissivity")
end

function request_alarm_inputs_info(withActive, fullList, onlyPort)
    local response, code = http_request("/cgi-bin/alarm.cgi?action=getInSlots")

    local result

    if string.is_empty(response) then
        return result, code
    end

    local digital_inputs_count = response:get_ini_value("result")

    local responseConfig, code = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=Alarm")

    if string.is_empty(responseConfig) then 
        return result, code
    end

    local portResult = json.object()

    for i, input_name in responseConfig:gmatch("table.Alarm%[(%d+)%]%.Name=%s*([^\r\n]+)") do

        if (fullList or (onlyPort == tostring(i))) then
            local port = json.object()
            local prefix = "Alarm["..tostring(i).."]."
            local input_name = responseConfig:get_ini_value(prefix.."Name")

            if not string.is_empty(input_name) then
                port = { name = input_name }
            end

            if withActive then
                local input_active = responseConfig:get_ini_value(prefix.."Enable")
                if not string.is_empty(input_active) and input_active == "true"  then
                    port.active = true
                else
                    port.active = false
                end
            end
            if next(port) ~= nil then
               portResult[tostring(i)] = port
            end
        end
    end

    if portResult ~= nil and next(portResult) ~= nil then
        result = portResult
    end

    return result, code
end

function setup:get_alarm_inputs_capability()

    if self.digital_inputs_ports ~= nil then
        return self.digital_inputs_ports
    end

    local result, code = request_alarm_inputs_info(false, true)

    if result ~= nil and next(result) ~= nil and code == http_request.OK_CODE then
        self.digital_inputs_ports =  result
        return self.digital_inputs_ports
    end

end

function setup_api_request_params(cdi_url)
    http_request:set_base_url(cdi_url or os.getenv("IVIDEON_BASE_URL"))
    if not http_request.base_url then error("Couldn't get api url") end

    http_request.timeout_ms = 10000
    http_request.max_retries_count = 1
end

function get_cgi_response_as_int(uri)
    local result = http_request(uri)
    if result ~= nil then
        local count = math.tointeger(result:get_ini_value("result"))
        if count ~= nil then return count end
        return 0
    end

    return nil -- request failed, maybe camera is anavailable
end

function check_mic_available()
    if settings.audio_input_ch_count == nil then
        settings.audio_input_ch_count = get_cgi_response_as_int("/cgi-bin/devAudioInput.cgi?action=getCollect") -- IPC only
    end

    if settings.audio_input_ch_count == nil or settings.audio_input_ch_count <= 0 then
        settings:remove("microphone/enabled")
        settings:remove("microphone/sensitivity")
        settings:remove("sound_detection/enabled")
        settings:remove("sound_detection/sensitivity")
        return false
    end
    return (settings.audio_input_ch_count > 0)
end

function check_speaker_available() -- req for push to talk
    if settings.audio_output_ch_count == nil then
        settings.audio_output_ch_count = get_cgi_response_as_int("/cgi-bin/devAudioOutput.cgi?action=getCollect")
        if settings.audio_output_ch_count == nil then
            warn_log("Failed to determine if the speaker is available")
            return false
        end
    end

    return (settings.audio_output_ch_count > 0)
end

function setup_settings_handlers()
    table.merge(settings.traits, {
        ["general/online_mode"] = {get = setup.get_online_mode_state, set = setup.set_online_mode_state},
        ["general/timezone"] = {get = setup.get_timezone, set = setup.set_timezone},
        ["night_vision/mode"] = {get = setup.get_ir_state, set = setup.set_ir_state},
        ["microphone/enabled"] = {get = setup.get_mic_state, set = setup.set_mic_state},
        ["sound_detection/enabled"] = {get = setup.get_sound_detector, set = setup.set_sound_detector},
        ["sound_detection/sensitivity"] = {get = setup.get_sound_detector, set = setup.set_sound_detector},
        ["microphone/sensitivity"] = {get = setup.get_mic_sensitivity, set = setup.set_mic_sensitivity},
        ["general/power_freq"] = {get = setup.get_power_freq, set = setup.set_power_freq},
        ["microphone/denoise"] = {get = setup.get_mic_denoise, set = setup.set_mic_denoise},
        ["pir_detection/enabled"] = {get = setup.get_pir_settings, set = setup.set_pir_settings}
    })
    for _, stream in ipairs({"0", "1"}) do for _, v in ipairs({"resolution", "bitrate", "fps", "codec", "acodec", "i_frame_interval", "variable_bitrate"}) do
        settings("streams/" .. stream .. "/" .. v).get = setup.get_streams_settings
        settings("streams/" .. stream .. "/" .. v).set = function(setup) return setup.set_stream_settings(setup, tonumber(stream)) end
    end end
    for _, v in ipairs({"datetime/enabled", "datetime/display_week", "datetime/24h", "datetime/format", "datetime/pos", "name/enabled", "name/text", "name/pos"}) do
        settings("osd/" .. v).get = setup.get_osd_settings ; settings("osd/" .. v).set = setup.set_osd_settings
    end
    for _, v in ipairs({"enabled", "0/sensitivity", "0/threshold", "0/zones"}) do
        settings("motion_detector/" .. v).get = setup.get_motion_detector
        if v == "0/zones" then settings("motion_detector/" .. v).set = setup.set_motion_detector_zone
        else settings("motion_detector/" .. v).set = setup.set_motion_detector end
    end

    for _, v in ipairs({"brightness", "contrast", "saturation", "hue"}) do
        settings("image/" .. v).get = setup.get_color_settings
        settings("image/" .. v).set = setup.set_color_settings
    end
    for _, v in ipairs({"upside_down", "rotate"}) do
        settings("image/" .. v).get = setup.get_videoin_settings
        settings("image/" .. v).set = setup.set_videoin_settings
    end
    for _, v in ipairs({"wdr", "wdr_level"}) do
        settings("image/" .. v).get = setup.get_videoin_backlight
        settings("image/" .. v).set = setup.set_videoin_backlight
    end
    for _, v in ipairs({"shutter_time", "gain"}) do
        settings("image/" .. v).get = setup.get_exposure_settings
        settings("image/" .. v).set = setup.set_exposure_settings
    end
    settings("image/wb").get = setup.get_wb_settings
    settings("image/wb").set = setup.set_wb_settings

    settings.traits.access_funcs_owner_table = setup
end

---------------------------------------
-- Device info functions
---------------------------------------

local function get_fw_version()
    local result = http_request("/cgi-bin/magicBox.cgi?action=getSoftwareVersion")
    if not result then
        return nil
    end

    local version = result:get_ini_value("version")
    if not version then
        return nil
    end

    if not version:lower():find("build", 1, true) then
        -- Version doesn't contain build data, trying to search it
        local build_date = result:get_ini_value("BuildDate")
        if build_date then
            version = version..",build:"..build_date
        end
    end

    return version
end

local function get_mac_address()
    local result, status = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=Network")
    if status ~= nil and status ~= 200 then
        return nil, status
    end

    if not result then
        return nil, 400
    end

    local default_iface = result:get_ini_value("table.Network.DefaultInterface")
    if not default_iface then
        return nil, 400
    end

    return result:get_ini_value("table.Network."..default_iface..".PhysicalAddress")
end

local function is_valid_zone(zone)
    return  zone.top_left[1] >= 0 and zone.top_left[1] < settings.cover.ZONE_BASE_WIDTH
            and zone.top_left[2] >= 0 and zone.top_left[2] < settings.cover.ZONE_BASE_HEIGHT
            and zone.bottom_right[1] >= 0 and zone.bottom_right[1] < settings.cover.ZONE_BASE_WIDTH
            and zone.bottom_right[2] >= 0 and zone.bottom_right[2] < settings.cover.ZONE_BASE_HEIGHT
            and zone.bottom_right[1] > zone.top_left[1]
            and zone.bottom_right[2] > zone.top_left[2]
end

local function get_privacy_masking()
    local result = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=VideoWidget")
    if not result then
        return nil
    end

    local covers_token = "table.VideoWidget%[" .. setup:ch() .. "%].Covers[^\r\n]+"
    local zones = {}
    for cover_line in result:gmatch(covers_token) do
        local zone_id = math.tointeger(cover_line:match("Covers%[(%d*)%]"))
        if not zone_id then
            return nil
        end

        if not zones[zone_id] then
            zones[zone_id] = {}
            zones[zone_id].top_left = {}
            zones[zone_id].bottom_right = {}
        end

        local zone_enabled = cover_line:get_ini_value("EncodeBlend")
        if zone_enabled then
            -- It's an EncodeBlend attribute
            zones[zone_id].enabled = (zone_enabled == "true")
        else
            local rect_point_type = math.tointeger(cover_line:match("Rect%[(%d*)%]"))
            if rect_point_type then
                -- no pre-escaping ('%') needed for key argument of get_ini_value(): it will be escaped automatically inside the function
                local point_coord = math.tointeger(cover_line:get_ini_value("Rect["..rect_point_type.."]"))
                if rect_point_type == 0 then -- left X
                    zones[zone_id].top_left[1] = point_coord
                elseif rect_point_type == 1 then -- top Y
                    zones[zone_id].top_left[2] = point_coord
                elseif rect_point_type == 2 then -- right X
                    zones[zone_id].bottom_right[1] = point_coord
                elseif rect_point_type == 3 then -- bottom Y
                    zones[zone_id].bottom_right[2] = point_coord
                end
            end
        end
    end

    -- Remove disabled and incorect zones
    for zone_id, zone in pairs(zones) do
        if not zone.enabled or not is_valid_zone(zone) then
            zones[zone_id] = nil
        end
    end

    if next(zones) then
        return {["base_width"] = settings.cover.ZONE_BASE_WIDTH, ["base_height"] = settings.cover.ZONE_BASE_HEIGHT, ["zones"] = zones}
    else
        -- There are no privacy zones
        return nil
    end
end

function get_relay_outputs_info()
    local result
    local response, code = http_request("cgi-bin/configManager.cgi?action=getConfig&name=AlarmOut")
    if not response then
        return result, code
    end

    local portResult = {}
    for port_num, mode in response:gmatch("table.AlarmOut%[(%d+)%]%.Mode=(%d+)") do
        -- mode is preserved for future usage
        portResult[tostring(port_num)] = {mode = "bistable"}
    end

    if next(portResult) ~= nil then
        result = portResult
    end

    return result, code
end

function setup:get_relay_outputs_capability()

    if self.relay_outputs_ports ~= nil then
        return self.relay_outputs_ports
    end

    local result, code = get_relay_outputs_info()

    if result ~= nil and next(result) ~= nil and code == http_request.OK_CODE then
        self.relay_outputs_ports = result
        return self.relay_outputs_ports
    end

end

function get_info_features()
    local features = json.object()

    local thermoConfig = setup:get_thermo_settings() -- If there is thermovision settings we'll append thermovision feature
    if not string.is_empty(thermoConfig) and tonumber(thermoConfig) ~= 0.0 then
        features["thermal"] = {
            ["provider"] = "dahua",
            ["version"] = "1.0",
            ["emission"] = tonumber(thermoConfig)
        }
    end

    local digitalInputsPorts = setup:get_alarm_inputs_capability() -- If there is alarm inputs settings we'll append digital_inputs feature
    if digitalInputsPorts ~= nil and next(digitalInputsPorts) ~= nil then
        features.digital_inputs = {
            ports = digitalInputsPorts,
            functions = {
                "device.ctl.digital_input.status"
            }
        }
        iv.log("Add Dahua Digital Input feature "..json.write(features.digital_inputs))
    end

    local relayOutputsPorts = setup:get_relay_outputs_capability() -- If there is alarm inputs settings we'll append relay_outputs feature
    if relayOutputsPorts ~= nil and next(relayOutputsPorts) ~= nil then
        features.relay_outputs = {
            ports = relayOutputsPorts,
            functions = {
                "device.ctl.relay_output.status"
            }
        }
        iv.log("Add Dahua Relay Outputs feature "..json.write(features.relay_outputs))
    end

    return features
end

function get_info()
    local device_info = json.object()
    device_info.serial_number = get_cam_property("getSerialNo", "sn")
    -- Most likely its wrong device so we shoudn't get all other fields
    if string.is_empty(device_info.serial_number) then return device_info end

    device_info.vendor = get_cam_property("getVendor", "vendor")
    device_info.model = get_cam_property("getDeviceType", "type")
    device_info.fw_version = get_fw_version()
    device_info.hardware_id = get_cam_property("getHardwareVersion", "version")
    device_info.mac_address = get_mac_address()
    device_info.privacy_masking = get_privacy_masking()
    device_info.serial_number_format = "dahua"
    device_info.features = get_info_features()

    return device_info
end

function get_max_extra_stream()
    local result, status = http_request("/cgi-bin/magicBox.cgi?action=getProductDefinition&name=MaxExtraStream")
    if type(result) == "string" and result:find("table.MaxExtraStream") then
        return result:get_ini_value("table.MaxExtraStream"), status
    end

    return nil, status
end

function get_rtsp_port()
    local result, status = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=RTSP")
    if type(result) == "string" and result:find("table.RTSP.Port") then
        return result:get_ini_value("table.RTSP.Port"), status
    end

    return nil, status
end

function probe(cmd)
    http_request.max_retries_count = 1

    if not cmd.args.inspection_protocol then
        cmd.args.inspection_protocol = cmd.args.cdi_scheme or "dahua"
    end

    local reply = cmd
    local props = {
        vendor = { name = "vendor name",      fn = function() return get_cam_property("getVendor", "vendor") end },
        model  = { name = "type",             fn = function() return get_cam_property("getDeviceType", "type") end },
        sn     = { name = "serial number",    fn = function() return get_cam_property("getSerialNo", "sn") end },
        mac    = { name = "mac address",      fn = function() return get_mac_address() end },
        sub    = { name = "max extra stream", fn = function() return get_max_extra_stream() end },
        rtsp   = { name = "rtsp port",        fn = function() return get_rtsp_port() end }
    }

    for _, fields in pairs(props) do
        local value, status = fields.fn()
        if status ~= nil and status ~= 200 then
            reply.args = { status = status,  reason = "failed to request device " .. fields.name }
            return reply
        end

        fields.value = value
    end

    local result = cmd.args
    result.manufacturer = props.vendor.value
    result.model = props.model.value
    result.description = props.model.value
    result.presentation_url = (cmd.args.use_https and "https://" or "http://") .. cmd.args.host .. ":" .. tostring(cmd.args.port)
    result.sn = props.sn.value
    result.mac = props.mac.value
    result.channels = { }

    local channels_list, status = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=ChannelTitle")
    if (not status or status == 200) and not channels_list then
        status = 400
    end

    if status ~= nil and status ~= 200 then
        reply.args = { status = status,  reason = "failed to request device channel list" }
        return reply
    end

    local rtsp_port = props.rtsp.value or "554"

    for i = 0, 64 do
        local title = channels_list:get_ini_value("table.ChannelTitle[" .. tostring(i) .. "].Name")
        if not title then break end

        local channel = { }
        local video_ch_idx = tostring(i + 1) -- the 'video channel index' differs from the index of ChannelTitle array by one (see 'ChannelTitle' request doc)
        channel.hw_id = video_ch_idx
        channel.video = {
            primary = http_request:copy_url_with{scheme = "rtsp", port = rtsp_port, path = "/cam/realmonitor", query = "channel=" .. video_ch_idx .. "&subtype=0"},
            preview = http_request:copy_url_with{path = "/cgi-bin/snapshot.cgi", query = "channel=" .. video_ch_idx},
            rtsp_transport = "tcp"
        }

        if props.sub.value then
            local sub = tonumber(props.sub.value)
            local names = { "secondary", "third" }
            for idx, name in ipairs(names) do
                if idx > sub then break end
                channel.video[name] = http_request:copy_url_with{scheme = "rtsp", port = rtsp_port, path = "/cam/realmonitor", query = "channel=" .. video_ch_idx .. "&subtype=" .. tostring(idx)}
            end
        end

        channel.events = "dahua?port=" .. tostring(cmd.args.port)
        channel.storage = http_request:copy_url_with{scheme = "dahua", query = (cmd.args.use_https and "https=true" or "")}
        channel.config = cmd.args.cdi_scheme .. "?port=" .. tostring(cmd.args.port)

        if not props.model.value:find("XVR") then
            local ptz_conf = http_request("/cgi-bin/configManager.cgi?action=getConfig&name=Ptz")
            if type(ptz_conf) == "string" and ptz_conf:find("ProtocolName") then
                channel.ptz = http_request:copy_url_with{scheme = "dahua-ptz", query = "channel=" .. video_ch_idx}
            else
                local ptz_test = http_request("/cgi-bin/devVideoInput.cgi?action=getFocusStatus")
                if type(ptz_test) == "string" and ptz_test:find("ZoomMotorSteps") then
                    channel.ptz = http_request:copy_url_with{scheme = "dahua-vf", query = "channel=" .. video_ch_idx}
                end
            end
        end

        result.channels[#result.channels + 1] = channel
    end

    reply.args = { status = 200,  result = result}
    return reply
end

-- Info command for the whole device
-- Note: if it's necessary to process also channel level info, then remove "level" field and process args.chIdx in the get_info() function
caps.requests.get_info = {
    useCache = true,
    level = "mod"
}
