require("cloud_config_template")

for _,v in ipairs({"led/enabled", "sound_detection/enabled", "sound_detection/sensitivity",
    "streams/2/resolution", "streams/2/bitrate", "streams/2/fps", "streams/2/i_frame_interval", "streams/2/variable_bitrate"}) do settings:remove_base(v) end

settings:add({
    { name = "pir_detection/enabled", meta = {type = "bool"} },
    { name = "motion_detector/0/sensitivity", meta = {type = "int", range = {0, 100, 20} } },
    -- { name = "motion_detector/0/threshold", meta = {type = "int", range = {0, 100, 20} } },
    { name = "motion_detector/0/zones", meta = {type = "grid", range = {22, 18} } }, --fixme
    { name = "microphone/denoise", meta = {type = "bool"} },
    { name = "osd/transparent", meta = {type = "bool"} },
    { name = "osd/flashing", meta = {type = "bool"} },
    { name = "osd/datetime/enabled", meta = {type = "bool"} },
    { name = "osd/datetime/display_week", meta = {type = "bool"} },
    { name = "osd/datetime/24h", meta = {type = "bool"} },
    { name = "osd/datetime/format", meta = {type = "choice", range = {} } },
    { name = "osd/datetime/pos", meta = {type = "position", range = {} } },
    { name = "osd/name/enabled", meta = {type = "bool"} },
    { name = "osd/name/text", meta = {type = "string", range = {0, 32}} },
    { name = "osd/name/pos", meta = {type = "position", range = {} } }
})

for i = 0, 5 do settings:add({
    { name = "osd/custom/" .. tostring(i) .. "/text", meta = {type = "string", range = {0, 32}} },
    { name = "osd/custom/" .. tostring(i) .. "/pos", meta = {type = "position", range = {} } },
    }) end

settings.sound_enabled = { true, true }

settings.traits = {
    ["streams/0/resolution"] = {
        depends = {"streams/0/fps", "streams/0/i_frame_interval", "streams/1/fps", "streams/1/i_frame_interval", "streams/1/resolution"},
        restart_requered = true
    },
    ["streams/1/resolution"] = {
        depends = {"streams/0/fps", "streams/0/i_frame_interval", "streams/1/fps", "streams/1/i_frame_interval"},
        restart_requered = true
    },
    ["streams/0/fps"] = {
        depends = {"streams/0/i_frame_interval"},
        restart_requered = true
    },
    ["streams/1/fps"] = {
        depends = {"streams/1/i_frame_interval"},
        restart_requered = true
    },
    ["streams/0/codec"] = {
        restart_requered = true
    },
    ["streams/1/codec"] = {
        restart_requered = true
    },
    ["streams/0/acodec"] = {
        restart_requered = true
    },
    ["streams/1/acodec"] = {
        restart_requered = true
    },
    ["microphone/enabled"] = {
        restart_requered = true
    },
    ["image/rotate"] = {
        restart_requered = true
    }
}

settings.vendor_name = "Hikvision"
settings.update_base_impl = settings.update

---------------------------------------
-- Common fucntions
---------------------------------------

function xml_request(anyUrl, method, requestData)
    local response_output, code = http_request(anyUrl, method, requestData)
    if response_output then
        local success, response = pcall( xml.dom.parseStr, response_output, {stripWhitespace=true} )
        if success == true then return response, code end
    end
    return nil, code
end

function xml.dom.testOrAddElements(parent, ...)
    local tt = parent
    for _, v in pairs({...}) do
        if type(tt) ~= "table" then return false end
        if type(tt[v]) ~= "table" then xml.dom.addElement(tt, v) end
        tt = tt[v]
    end
    return true
end

---------------------------------------
-- Setup functions
---------------------------------------

setup = {
    schedulerUrl = "",
    channelsUrl = "",
    ircutUrl = "",
    timeUrl = "/ISAPI/System/time",
    channelUrl = "",
    osdUrl = "",
    pirUrl = "/ISAPI/WLAlarm/PIR",
    mdUrl = "",
    thermoUrl = "/ISAPI/Thermal",
    anprUrl = "/ISAPI/ITC",
    acsUrl = "/ISAPI/AccessControl",
    acsUserUrl = "/ISAPI/AccessControl/UserInfo",
    acsCardUrl = "/ISAPI/AccessControl/CardInfo",
    acsFaceLibUrl = "/ISAPI/Intelligent/FDLib",
    acsFingerprintUrl = "/ISAPI/AccessControl/FingerPrintCfg",
    digitalIOUrl = "/ISAPI/System/IO",
    digitalInputsUrl = "/ISAPI/System/IO/inputs",
    cameraId = "",
    isProxied = nil,
    substreamCount = 2
}

function setup:set_camera(camera)
    local cam_info = vs.get_camera_info( tonumber(camera) )

    if log_config.tag == tostring(camera) and type(settings["general/online_mode"]) == "table" then return end -- already set
    log_config.tag = tostring(camera)
    if type(cam_info) ~= "table" or string.is_empty(cam_info.urlHigh) then return end


    local cam_id = "1"
    local parsedUrl = iv.parse_url(cam_info.urlHigh)
    if not string.is_empty(parsedUrl.path) then
        local url_parts = parsedUrl.path:split("/")
        cam_id = url_parts[table.getn(url_parts)]
        if #cam_id > 2 then cam_id = cam_id:sub(1,-3) end
    end

    if setup.cameraId == cam_id then return end
    setup.cameraId = cam_id

    local substream_count = racm_substream_count(cam_id)
    setup.substreamCount = substream_count
    if substream_count < 2 then
        settings:remove("streams/1/fps")
        settings:remove("streams/1/bitrate")
        settings:remove("streams/1/resolution")
        settings:remove("streams/1/codec")
        settings:remove("streams/1/acodec")
        settings:remove("streams/1/i_frame_interval")
        settings:remove("streams/1/variable_bitrate")
    end

    setup.schedulerUrl = string.format("/ISAPI/ContentMgmt/record/tracks/%s01", cam_id)

    setup.isProxied = (substream_count < 4)
    if setup.isProxied then
        setup.channelsUrl = "/ISAPI/ContentMgmt/StreamingProxy"
        setup.imageUrl = "/ISAPI/ContentMgmt/ImageProxy"
        setup.mdUrl = string.format("/ISAPI/ContentMgmt/InputProxy/channels/%s/video/motionDetection", cam_id)
        setup.channelUrl = string.format("/ISAPI/ContentMgmt/InputProxy/channels/%s", cam_id)
        setup.osdUrl = string.format("/ISAPI/ContentMgmt/InputProxy/channels/%s/video/overlays", cam_id)
    else
        setup.channelsUrl = "/ISAPI/Streaming"
        setup.imageUrl = "/ISAPI/Image"
        setup.mdUrl = string.format("/ISAPI/System/Video/inputs/channels/%s/motionDetection", cam_id)
        setup.channelUrl = string.format("/ISAPI/System/Video/inputs/channels/%s", cam_id)
        setup.osdUrl = string.format("/ISAPI/System/Video/inputs/channels/%s/overlays", cam_id)
    end

    setup.channelsUrl = setup.channelsUrl .. string.format("/channels/%s0", cam_id) -- NOTE: Without substream id!
    setup.imageUrl = setup.imageUrl .. string.format("/channels/%s", cam_id)
    setup.ircutUrl = setup.imageUrl .. "/ircutFilter"

    settings:load_from_storage( tonumber(camera) )
end

function setup:get_online_mode_state()
    local track_settings, _ = xml_request(self.schedulerUrl)
    if track_settings == nil then return end

    if track_settings.Track == nil then return end

    settings["general/online_mode"].value = track_settings.Track.CustomExtensionList.CustomExtension.enableSchedule["#text"] == "true"
    return true
end

function setup:set_mute_server_state(newState)
    local curState = settings["general/online_mode"].value and "off" or "on"
    if curState == newState then return end

    local request = (newState == "on") and "false" or "true"
    local track_settings , _ = xml_request(self.schedulerUrl)
    if track_settings == nil then return end
    track_settings.Track.CustomExtensionList.CustomExtension.enableSchedule["#text"] = request

    request = xml.write(track_settings, {pretty=false, escape=false})
    local response, _ = xml_request(self.schedulerUrl, "PUT", request)

    assert(response.ResponseStatus.statusString["#text"] == "OK")
    settings["general/online_mode"].value = (newState ~= "on")
end

function setup:save_online_mode()
    local do_enable_schedule = tostring(settings["general/online_mode"].value)
    local track_settings, _ = xml_request(self.schedulerUrl)
    if track_settings == nil then return false end
    track_settings.Track.CustomExtensionList.CustomExtension.enableSchedule["#text"] = do_enable_schedule

    request = xml.write(track_settings, {pretty=false, escape=false})

    local response, _ = xml_request(self.schedulerUrl, "PUT", request)
    if response ~= nil and response.ResponseStatus.statusString["#text"] == "OK" then return true end
    return false
end

function setup:is_sound_turned_off()
    local audio = settings["microphone/enabled"]
    if audio then
        return not audio.value
    else
        return true
    end
end

-- Turn sound on stream streamId off if turn_off = true
function setup:turn_stream_sound_off(streamId, turn_off)
    if turn_off == self:is_sound_turned_off() then return false end
    local do_enable_sound = tostring( not turn_off )

    local request = "<StreamingChannel><Audio><enabled>" .. do_enable_sound .. "</enabled></Audio></StreamingChannel>"
    for i = 1, 5 do
        local response, _ = xml_request(self.channelsUrl .. streamId, "PUT", request)
        if response ~= nil and response.ResponseStatus.statusString["#text"] == "OK" then return true end
        iv.sleep(1000)
    end

    return false
end

function setup:set_mute_sound(newState)
    if not settings["microphone/enabled"] then return false end

    local return_val = {is_stream_restarted = false}
    local do_turn_sound_off = false
    if newState == "on" then do_turn_sound_off = true end

    for i = 1, setup.substreamCount do
        if self:turn_stream_sound_off(i, do_turn_sound_off) then
            return_val.is_stream_restarted = true
            if (settings["microphone/enabled"] ~= nil) then
                settings.sound_enabled[i] = (do_turn_sound_off == false)
                settings["microphone/enabled"].value = settings.sound_enabled[1] and settings.sound_enabled[2]
            end
        end
    end

    return return_val
end

function setup:get_irled_state()
    if not settings["night_vision/mode"] then return false end
    local response, _ = xml_request(self.ircutUrl)
    if not table.is_subtable_seq(response, "IrcutFilter", "IrcutFilterType") then
        settings:remove("night_vision/mode")
        return false -- TODO remove this plugin
    end

    response = response.IrcutFilter.IrcutFilterType["#text"]
    if response == "night" then
        settings["night_vision/mode"].value = "on"
    elseif response == "day" then
        settings["night_vision/mode"].value = "off"
    else
        settings["night_vision/mode"].value = "auto"
    end

    return settings["night_vision/mode"].value
end

function setup:set_irled(newState)
    if settings["night_vision/mode"].value == nil then return false end

    local set_ir_mode = "auto"
    if newState == "on" then
        set_ir_mode = "night"
    elseif newState == "off" then
        set_ir_mode = "day"
    end
    local request = "<?xml version=1.0 encoding=UTF-8?><IrcutFilter><IrcutFilterType>" .. set_ir_mode .. "</IrcutFilterType></IrcutFilter>"

    response, _ = xml_request(self.ircutUrl, "PUT", request)
    if response == nil then return false end
    if (response.ResponseStatus.statusString["#text"] == "OK") then
        settings["night_vision/mode"].value = newState
        return true
    else
        return false
    end
end

function setup:save_irled()
    return self:set_irled(settings["night_vision/mode"].value)
end

function setup:load_stream_settings(stream)
    local base_path = "streams/" .. tostring(stream) .. "/"
    local cur_settings, _ = xml_request(self.channelsUrl .. tostring(stream + 1) .. "/capabilities")
    local cur_settings_values, _ = xml_request(self.channelsUrl .. tostring(stream + 1))
    -- TODO: dynamicCap with parameters:
    if not table.is_subtable_seq(cur_settings, "StreamingChannel", "Video") then
        if table.is_subtable_seq(cur_settings, "ResponseStatus", "subStatusCode") then
            settings:remove(base_path .. "resolution")
            settings:remove(base_path .. "bitrate")
            settings:remove(base_path .. "fps")
            settings:remove(base_path .. "i_frame_interval")
            settings:remove(base_path .. "variable_bitrate")
            settings:remove(base_path .. "codec")
            settings:remove(base_path .. "acodec")
            settings:remove("microphone/enabled")
        end
        return
    end

    if not settings[base_path .. "resolution"] then return end -- triple-stream unsupported yet!

    local v = cur_settings.StreamingChannel.Video
    local v_values = cur_settings_values.StreamingChannel.Video

    settings[base_path .. "resolution"].value = v_values.videoResolutionWidth["#text"] .. "x" .. v_values.videoResolutionHeight["#text"]
    local r = {}
    local res_widths, res_heights = v.videoResolutionWidth["#attr"].opt:split(","), v.videoResolutionHeight["#attr"].opt:split(",")
    for i = 1, #res_heights do r[i] = tostring(res_widths[i]) .. "x" .. tostring(res_heights[i]) end
    settings[base_path .. "resolution"].meta.range = r

    local max_bitrate = 32*1024
    local min_bitrate = 0
    local qct = "VBR"
    if table.is_subtable_seq(v_values, "videoQualityControlType") then
        qct = v_values.videoQualityControlType["#text"]
        settings[base_path .. "variable_bitrate"].value = (qct == "VBR")
    else
        settings:remove(base_path .. "variable_bitrate")
    end

    if (qct == "VBR") then
        settings[base_path .. "bitrate"].value = tonumber(v_values.vbrUpperCap["#text"])
        min_bitrate = tonumber(v.vbrUpperCap["#attr"].min or min_bitrate)
        max_bitrate = tonumber(v.vbrUpperCap["#attr"].max or max_bitrate)
    else
        settings[base_path .. "bitrate"].value = tonumber(v_values.constantBitRate["#text"])
        min_bitrate = tonumber(v.constantBitRate["#attr"].min or min_bitrate)
        max_bitrate = tonumber(v.constantBitRate["#attr"].max or max_bitrate)
    end

    local bitrate_values = {128, 256, 512, 1024, 2048, 3072, 4096, 6144, 8192}
    local rate_cur_range = {}
    for i = 1, #bitrate_values do
        if bitrate_values[i] >= min_bitrate and bitrate_values[i] <= max_bitrate then
            rate_cur_range[#rate_cur_range+1] = bitrate_values[i]
        end
    end
    if not is_in_table(settings[base_path .. "bitrate"].value, rate_cur_range) then
        rate_cur_range[#rate_cur_range+1] = settings[base_path .. "bitrate"].value
    end
    settings[base_path .. "bitrate"].meta.range = rate_cur_range

    settings[base_path .. "fps"].value = math.floor(tonumber(v_values.maxFrameRate["#text"]) / 100)
    local fps_values = v.maxFrameRate["#attr"].opt:split(",")
    local fps_cur_range = {}
    for i = 1, #fps_values do
        if tonumber(fps_values[i]) >= 100 then -- remove FPS values less 1 FPS
            fps_cur_range[#fps_cur_range+1] = math.floor(tonumber(fps_values[i]) / 100)
        end
    end
    settings[base_path .. "fps"].meta.range = fps_cur_range
    if not is_in_table(settings[base_path .. "fps"].value, settings[base_path .. "fps"].meta.range) then
        settings[base_path .. "fps"].value = settings[base_path .. "fps"].meta.range[1]
    end

    local actual_gop = nil
    if type(v_values.GovLength) == "table" then old_gop = tonumber(v_values.GovLength["#text"])
    elseif type(v_values.IFrameInterval) == "table" then old_gop = tonumber(v_values.IFrameInterval["#text"]) end
    if actual_gop then
        settings[base_path .. "i_frame_interval"].value = actual_gop
        settings[base_path .. "i_frame_interval"].meta.range = {1, settings[base_path .. "fps"].value * 4}
    else settings:remove(base_path .. "i_frame_interval") end

    -- min_gov = tonumber(v.GovLength["#attr"].min or 1)
    -- max_gov = tonumber(v.GovLength["#attr"].max or settings[base_path .. "fps"].value * 4)
    -- v.keyFrameInterval / min=40 max=16000 #text=2000 in ms between P-frames !!

    if settings[base_path .. "codec"] then
        settings[base_path .. "codec"].value = v_values.videoCodecType["#text"]
        local r = {}
        local vcodecs = v.videoCodecType["#attr"].opt:split(",")
        for i = 1, #vcodecs do
            if not vcodecs[i]:find("JP") then
                table.insert(r, vcodecs[i])
            end
        end
        if #r > 1 then settings[base_path .. "codec"].meta.range = r
        else settings:remove(base_path .. "codec") end
    end

    local a = cur_settings.StreamingChannel.Audio
    local a_values = cur_settings_values.StreamingChannel.Audio
    if a_values ~= nil and a ~= nil and settings[base_path .. "acodec"] then
        settings[base_path .. "acodec"].value = a_values.audioCompressionType["#text"]
        local r = {}
        local acodecs = a.audioCompressionType["#attr"].opt:split(",")
        for i = 1, #acodecs do r[i] = acodecs[i] end
        if #r > 1 then settings[base_path .. "acodec"].meta.range = r
        else settings:remove(base_path .. "acodec") end
    else
        settings:remove(base_path .. "acodec")
    end

    if a_values ~= nil and settings["microphone/enabled"] then
        -- Make [1 stream] AND [2 stream] generic setting
        settings.sound_enabled[stream] = (a_values.enabled["#text"] == "true")
        settings["microphone/enabled"].value = settings.sound_enabled[1] and settings.sound_enabled[2]
    else
        settings:remove("microphone/enabled")
    end
end

function setup:load_streams_settings()
    for stream = 1, setup.substreamCount do
        self:load_stream_settings(stream - 1)
    end
end

function setup:set_stream_settings(stream, force_changes)
    local is_changed = force_changes or false
    local cur_settings, _ = xml_request(self.channelsUrl .. tostring(stream + 1))
    if not table.is_subtable_seq(cur_settings, "StreamingChannel", "Video") then
        err_log("Can't get current settings")
        return false
    end

    local v = cur_settings.StreamingChannel.Video
    local base_path = "streams/" .. tostring(stream) .. "/"

    if settings[base_path .. "resolution"] == nil or settings[base_path .. "resolution"].value == nil then return false end
    local resolution = settings[base_path .. "resolution"].value:split("x")
    if v.videoResolutionWidth["#text"] ~= resolution[1] or v.videoResolutionHeight["#text"] ~= resolution[2] then
        is_changed = true
        v.videoResolutionWidth["#text"], v.videoResolutionHeight["#text"] = resolution[1], resolution[2]
    end

    local new_fps = tostring( math.floor(tonumber(settings[base_path .. "fps"].value) * 100) )
    if v.maxFrameRate["#text"] ~= new_fps then
        is_changed = true
        v.maxFrameRate["#text"] = new_fps
    end

    local new_gop = (settings[base_path .. "i_frame_interval"] and settings[base_path .. "i_frame_interval"].changed) and settings[base_path .. "i_frame_interval"].value or tonumber(settings[base_path .. "fps"].value) * 2
    local old_gop = new_gop
    if type(v.GovLength) == "table" then old_gop = tonumber(v.GovLength["#text"])
    elseif type(v.IFrameInterval) == "table" then old_gop = tonumber(v.IFrameInterval["#text"]) end
    if old_gop ~= new_gop then
        is_changed = true
        if type(v.GovLength) == "table" then
            v.GovLength["#text"] = tostring(new_gop)
        elseif type(v.IFrameInterval) == "table" then
            v.IFrameInterval["#text"] = tostring(new_gop)
        else
            xml.dom.addElement(v, "GovLength", tostring(new_gop))
        end
    end

    if type(v.keyFrameInterval) == "table" then v.keyFrameInterval["#text"] = math.floor(1000 * (new_gop / tonumber(settings[base_path .. "fps"].value) )) end

    local qct = "VBR"
    if settings[base_path .. "variable_bitrate"] then
        qct = (settings[base_path .. "variable_bitrate"].value) and "VBR" or "CBR"
        if table.is_subtable_seq(v, "videoQualityControlType") and v.videoQualityControlType["#text"] ~= qct then
            is_changed = true
            v.videoQualityControlType["#text"] = qct
        else
            is_changed = true
            xml.dom.addElement(v, "videoQualityControlType", qct)
        end
    end

    if v.constantBitRate ~= nil and v.constantBitRate["#text"] ~= tostring(settings[base_path .. "bitrate"].value) then
        is_changed = true
        v.constantBitRate["#text"] = tostring(settings[base_path .. "bitrate"].value)
    end
    if v.vbrUpperCap ~= nil and v.vbrUpperCap["#text"] ~= tostring(settings[base_path .. "bitrate"].value) then
        is_changed = true
        v.vbrUpperCap["#text"] = tostring(settings[base_path .. "bitrate"].value)
    end

    if settings[base_path .. "codec"] then
        local codec = settings[base_path .. "codec"].value
        if v.videoCodecType["#text"] ~= codec then
            is_changed = true
            v.videoCodecType["#text"] = codec
        end
    end

    local a = cur_settings.StreamingChannel.Audio
    if a ~= nil then
        if a.enabled ~= nil and settings["microphone/enabled"] ~= nil then
            if a.enabled["#text"] ~= tostring(settings["microphone/enabled"].value) then
                is_changed = true
                a.enabled["#text"] = tostring(settings["microphone/enabled"].value)
            end
        end

        if a.audioCompressionType ~= nil and settings[base_path .. "acodec"] ~= nil then
            if a.audioCompressionType["#text"] ~= settings[base_path .. "acodec"].value then
                is_changed = true
                a.audioCompressionType["#text"] = settings[base_path .. "acodec"].value
            end
        end
    end

    if is_changed then
        local req = xml.write(cur_settings)
        for tries = 1, 5 do
            local response, _ = xml_request(self.channelsUrl .. tostring(stream + 1), "PUT", req)
            if response ~= nil and response.ResponseStatus.statusString["#text"] == "OK" then
                return true
            end

            if not v.videoScanType and response ~= nil and response.ResponseStatus.subStatusCode["#text"] == "badXmlContent" then
                xml.dom.addElement(v, "videoScanType", "progressive")
                req = xml.write(cur_settings)
            end

            iv.sleep(2000)
        end
    else
        return true
    end

    return false
end

function setup:set_streams_settings()
    local success = false
    for stream = 1, setup.substreamCount do
        if self:set_stream_settings(stream - 1) then
            success = true
        end
    end

    return success
end

-- Hikvision used broken CST timezone notation: CST[+-]HH:MM:SS with INVERTED sign = UTC[-+]HH:MM
function setup:get_timezone()
    local response, _ = xml_request(self.timeUrl)
    if response ~= nil then
        response = response.Time.timeZone["#text"] --:split("+-")[2]
        local sign, hour, minute = response:match("([-+])(%d+):?(%d+):?(%d*)(.*)$")
        sign = (sign == "-") and "+" or "-"
        hour = (#hour < 2) and "0" .. hour or hour
        settings["general/timezone"].value = sign .. hour .. ":" .. minute
    end

    if settings["general/timezone"].meta.range then return end
    local r = {}
    special = {"-04:30", "-03:30", "+03:30", "+04:30", "+05:30", "+05:45", "+06:30", "+09:30"}
    for i = -12, 13 do
        local tz = ""
        if i >= 0 then tz = tz .. "+" else tz = tz .. "-" end
        if math.abs(i) < 10 then tz = tz .. "0" end
        tz = tz .. tostring(math.abs(i)) .. ":00"
        r[#r+1] = tz
        for _, v in ipairs(special) do if v:sub(1,3) == tz:sub(1,3) then r[#r+1] = v end end
    end
    settings["general/timezone"].meta.range = r
end

function setup:set_timezone()
    local sign, hour, minute = settings["general/timezone"].value:match("([-+])(%d+):?(%d+):?(%d*)$")
    sign = (sign == "-") and "+" or "-"
    hour = (hour:sub(1,1) == "0") and hour:sub(2,2) or hour
    local newTimezone = "CST" .. sign .. hour .. ":" .. minute .. ":00"
    local response, _ = xml_request(self.timeUrl, "PUT", "<Time><timeMode>NTP</timeMode><timeZone>" .. newTimezone .. "</timeZone></Time>")

    if response ~= nil and response.ResponseStatus.statusString["#text"] == "OK" then return true end
    return false
end

function setup:get_osd_settings()
    local caps, _ = xml_request(self.osdUrl .. "/capabilities")
    local cs, _ = xml_request(self.osdUrl)
    local pos_range = {704, 576}
    if table.is_subtable_seq(cs, "VideoOverlay", "normalizedScreenSize", "normalizedScreenWidth") then
        pos_range = {
            tonumber(cs.VideoOverlay.normalizedScreenSize.normalizedScreenWidth["#text"]),
            tonumber(cs.VideoOverlay.normalizedScreenSize.normalizedScreenHeight["#text"])
        }
    else
        return
    end

    local attr = caps.VideoOverlay.attribute
    if not attr then attr = caps.VideoOverlay end -- IPC hack (not documented)
    if table.is_subtable_seq(attr, "transparent") and #(attr.transparent["#attr"].opt:split(",")) > 1 then
        settings["osd/transparent"].value = (attr.transparent["#text"] == "true")
    else
        settings:remove("osd/transparent")
    end

    if table.is_subtable_seq(attr, "flashing") and #(attr.flashing["#attr"].opt:split(",")) > 1 then
        settings["osd/flashing"].value = (attr.flashing["#text"] == "true")
    else
        settings:remove("osd/flashing")
    end

    local datetime = caps.VideoOverlay.DateTimeOverlay
    if settings["osd/datetime/enabled"] and table.is_subtable_seq(datetime, "enabled") then
        settings["osd/datetime/enabled"].value = (datetime.enabled["#text"] == "true")
    else
        settings:remove("osd/datetime/enabled")
        settings:remove("osd/datetime/pos")
        settings:remove("osd/datetime/format")
        settings:remove("osd/datetime/24h")
        settings:remove("osd/datetime/display_week")
    end
    if settings["osd/datetime/pos"] and table.is_subtable_seq(datetime, "positionX") then
        settings["osd/datetime/pos"].meta.range = pos_range
        settings["osd/datetime/pos"].value = {
            tonumber(datetime.positionX["#text"]),
            tonumber(datetime.positionY["#text"])
        }
    else
        settings:remove("osd/datetime/pos")
    end

    if settings["osd/datetime/format"] and table.is_subtable_seq(datetime, "dateStyle") then
        settings["osd/datetime/format"].value = datetime.dateStyle["#text"]

        local format_variants = datetime.dateStyle["#attr"].opt:split(",")
        local r = {}
        for i = 1, #format_variants do r[i] = format_variants[i] end -- -- doc 8.4.25
        if #format_variants < 2 then settings:remove("osd/datetime/format") end
        settings["osd/datetime/format"].meta.range = r
    else
        settings:remove("osd/datetime/format")
    end
    if settings["osd/datetime/24h"] and table.is_subtable_seq(datetime, "timeStyle") then
        settings["osd/datetime/24h"].value = (datetime.timeStyle["#text"] == "24hour")
        local style_variants =  datetime.timeStyle["#attr"].opt:split(",")
        if #style_variants < 2 then settings:remove("osd/datetime/24h") end
    else
        settings:remove("osd/datetime/24h")
    end
    if settings["osd/datetime/display_week"] and table.is_subtable_seq(datetime, "displayWeek") then
        settings["osd/datetime/display_week"].value = (datetime.displayWeek["#text"] == "true")
        local week_variants =  datetime.displayWeek["#attr"].opt:split(",")
        if #week_variants < 2 then settings:remove("osd/datetime/display_week") end
    else
        settings:remove("osd/datetime/display_week")
    end

    local name = caps.VideoOverlay.channelNameOverlay
    if settings["osd/name/enabled"] and table.is_subtable_seq(name, "enabled") then
        settings["osd/name/enabled"].value = (name.enabled["#text"] == "true")
    else
        settings:remove("osd/name/enabled")
        settings:remove("osd/name/pos")
        settings:remove("osd/name/text")
    end
    if settings["osd/name/pos"] and table.is_subtable_seq(name, "positionX") then
        settings["osd/name/pos"].meta.range = pos_range
        settings["osd/name/pos"].value = {
            tonumber(name.positionX["#text"]),
            tonumber(name.positionY["#text"])
        }
    else
        settings:remove("osd/name/pos")
    end

    local customs = 0
    if table.is_subtable_seq(caps, "VideoOverlay", "TextOverlayList") then
        local overlays = caps.VideoOverlay.TextOverlayList
        customs = tonumber(overlays["#attr"].size or 0)
        if setup.isProxied then customs = 1 end -- All cameras can't setup more than 1 OSD custom text
        if customs > 5 then customs = 5 end

        if table.is_subtable_seq(overlays, "TextOverlay") then overlays = overlays.TextOverlay else overlays = {} end
        if #overlays == 0 then overlays = {overlays} end -- force array
        for i = 1, customs do
            if overlays[i] and overlays[i].displayText then
                settings["osd/custom/" .. tostring(i - 1) .. "/text"].value = overlays[i].displayText["#text"]
            else
                settings["osd/custom/" .. tostring(i - 1) .. "/text"].value = ""
            end
            local pos = settings["osd/custom/" .. tostring(i - 1) .. "/pos"]
            if overlays[i] and overlays[i].positionX then
                pos.value = {
                    tonumber(overlays[i].positionX["#text"]),
                    tonumber(overlays[i].positionY["#text"])
                }
            else
                pos.value = {0, 0}
            end
            pos.meta.range = pos_range
        end
    end
    for i = customs, 5 do
        settings:remove("osd/custom/" .. tostring(i) .. "/text")
        settings:remove("osd/custom/" .. tostring(i) .. "/pos")
    end
end

function setup:set_osd_settings()
    local cs, _ = xml_request(self.osdUrl)
    local caps, _ = xml_request(self.osdUrl .. "/capabilities")
    if not table.is_subtable_seq(cs, "VideoOverlay") or not table.is_subtable_seq(caps, "VideoOverlay") then return false end
    caps = caps.VideoOverlay

    local function set_xml_setting(cloud_setting, hik_name, caps_xml, req_subxml)
        if table.is_subtable_seq(req_subxml, hik_name) then
            if not cloud_setting then return end
            req_subxml[hik_name]["#text"] = tostring(cloud_setting.value)
            local variants = caps_xml[hik_name]["#attr"].opt
            if type(variants) == "string" then
                variants = variants:split(",")
                if #variants <= 1 or not is_in_table(req_subxml[hik_name]["#text"], variants) then req_subxml[hik_name]["#text"] = variants[1] end
            end
        end
    end

    local attr = cs.VideoOverlay.attribute
    set_xml_setting(settings["osd/transparent"], "transparent", caps.attribute or caps, attr)
    set_xml_setting(settings["osd/flashing"], "flashing", caps.attribute or caps, attr)

    local datetime = cs.VideoOverlay.DateTimeOverlay

    set_xml_setting(settings["osd/datetime/enabled"], "enabled", caps.DateTimeOverlay, datetime)
    if settings["osd/datetime/pos"] and table.is_subtable_seq(datetime, "positionX") then
        datetime.positionX["#text"] = tostring(settings["osd/datetime/pos"].value[1])
        datetime.positionY["#text"] = tostring(settings["osd/datetime/pos"].value[2])
    end
    set_xml_setting(settings["osd/datetime/display_week"], "displayWeek", caps.DateTimeOverlay, datetime)
    set_xml_setting(settings["osd/datetime/format"], "dateStyle", caps.DateTimeOverlay, datetime)
    local fake_cc_setting = {value = "12hour"}
    if settings["osd/datetime/24h"] and settings["osd/datetime/24h"].value then fake_cc_setting.value = "24hour" end
    set_xml_setting(fake_cc_setting, "timeStyle", caps.DateTimeOverlay, datetime)

    local name = cs.VideoOverlay.channelNameOverlay
    set_xml_setting(settings["osd/name/enabled"], "enabled", caps.channelNameOverlay, name)
    if settings["osd/name/pos"] and table.is_subtable_seq(name, "positionX") then
        name.positionX["#text"] = tostring(settings["osd/name/pos"].value[1])
        name.positionY["#text"] = tostring(settings["osd/name/pos"].value[2])
    end
    cs = xml.write(cs, {pretty=false})

    local response, _ = xml_request(self.osdUrl, "PUT", cs)
    if response == nil or response.ResponseStatus.statusString["#text"] ~= "OK" then return false end
    return true
end

function setup:get_image_settings()
    -- Fill setting with bool/int/choice automatically from cap
    local function fill_options(s_name, s_xml)
        local iv_s = settings[s_name]
        if not iv_s or not s_xml then
            settings:remove(s_name)
            return
        end
        if iv_s.meta.type ~= "bool" then
            iv_s.value = tonumber(s_xml["#text"]) or s_xml["#text"] -- int or string
        else
            iv_s.value = (s_xml["#text"] == "true") -- bool
        end

        if (s_xml["#attr"].min ~= nil) then
            iv_s.meta.type = "int"
            iv_s.meta.range = {tonumber(s_xml["#attr"].min), tonumber(s_xml["#attr"].max)}
        elseif (s_xml["#attr"].opt ~= nil) then
            local opts = s_xml["#attr"].opt:split(",")
            if #opts < 2 then
                settings:remove(s_name)
                return
            end

            if iv_s.meta.type ~= "bool" then
                local r = {}
                for i = 1, #opts do r[#r + 1] = opts[i] end
                if not is_in_table(iv_s.value, r) then r[#r + 1] = iv_s.value end
                iv_s.meta.type = "choice"
                iv_s.meta.range = r
            end
        else
            settings:remove(s_name)
        end
    end

    local caps, _ = xml_request(self.imageUrl .. "/capabilities")
    -- local cs = xml_request(self.imageUrl)
    if not table.is_subtable_seq(caps, "ImageChannel", "Color") then
        settings:remove("image/brightness")
        settings:remove("image/contrast")
        settings:remove("image/saturation")
        settings:remove("image/hue")
    else
        local color = caps.ImageChannel.Color
        fill_options("image/brightness", color.brightnessLevel)
        fill_options("image/contrast", color.contrastLevel)
        fill_options("image/saturation", color.saturationLevel)
        fill_options("image/hue", color.hueLevel)
    end

    if table.is_subtable_seq(caps, "ImageChannel", "corridor", "enabled") then
        fill_options("image/rotate", caps.ImageChannel.corridor.enabled)
    else settings:remove("image/rotate") end

    if table.is_subtable_seq(caps, "ImageChannel", "ImageFlip", "enabled") then
        fill_options("image/upside_down", caps.ImageChannel.ImageFlip.enabled)
    else settings:remove("image/upside_down") end

    if table.is_subtable_seq(caps, "ImageChannel", "WhiteBalance", "WhiteBalanceStyle") then
        fill_options("image/wb", caps.ImageChannel.WhiteBalance.WhiteBalanceStyle)
    else settings:remove("image/wb") end

    if table.is_subtable_seq(caps, "ImageChannel", "powerLineFrequency", "powerLineFrequencyMode") then
        fill_options("general/power_freq", caps.ImageChannel.powerLineFrequency.powerLineFrequencyMode)
    else settings:remove("general/power_freq") end

    if table.is_subtable_seq(caps, "ImageChannel", "Gain", "GainLevel") then
        fill_options("image/gain", caps.ImageChannel.Gain.GainLevel)
    else settings:remove("image/gain") end

    if table.is_subtable_seq(caps, "ImageChannel", "Exposure", "ExposureType") and table.is_subtable_seq(caps, "ImageChannel", "Shutter", "ShutterLevel") then
        local exp = caps.ImageChannel.Exposure.ExposureType
        local sl = caps.ImageChannel.Shutter.ShutterLevel
        if exp["#text"] == "manual" then settings["image/shutter_time"].value = sl["#text"]
        else settings["image/shutter_time"].value = "auto" end

        local sls = sl["#attr"].opt:split(",")
        local exps = exp["#attr"].opt:split(",")
        local r = {}
        if is_in_table("auto", exps) then r[#r + 1] = "auto" end
        for i = 1, #sls do r[#r + 1] = sls[i] end
        settings["image/shutter_time"].meta.range = r
    else settings:remove("image/shutter_time") end
end

function setup:set_image_settings()
    local cs, _ = xml_request(self.imageUrl)
    local ic = cs.ImageChannel -- throw
    if table.is_subtable_seq(ic, "Color") then
        local color = ic.Color
        if settings["image/brightness"] then color.brightnessLevel["#text"] = tostring(settings["image/brightness"].value) end
        if settings["image/contrast"] then color.contrastLevel["#text"] = tostring(settings["image/contrast"].value) end
        if settings["image/saturation"] then color.saturationLevel["#text"] = tostring(settings["image/saturation"].value) end
        if settings["image/hue"] then color.hueLevel["#text"] = tostring(settings["image/hue"].value) end
    end

    if table.is_subtable_seq(ic, "WhiteBalance", "WhiteBalanceStyle") and settings["image/wb"] then
        ic.WhiteBalance.WhiteBalanceStyle["#text"] = settings["image/wb"].value
    end

    if table.is_subtable_seq(ic, "powerLineFrequency", "powerLineFrequencyMode") and settings["general/power_freq"] then
        ic.powerLineFrequency.powerLineFrequencyMode["#text"] = settings["general/power_freq"].value
    end

    if table.is_subtable_seq(ic, "Exposure", "ExposureType") and table.is_subtable_seq(ic, "Shutter", "ShutterLevel") and settings["image/shutter_time"] then
        local exp_type = "manual"
        if settings["image/shutter_time"].value == "auto" then
            exp_type = "auto"
        else
            ic.Shutter.ShutterLevel["#text"] = settings["image/shutter_time"].value
        end
        ic.Exposure.ExposureType["#text"] = exp_type
    end

    if table.is_subtable_seq(ic, "Gain", "GainLevel") and settings["image/gain"] then
        ic.Gain.GainLevel["#text"] = settings["image/gain"].value
    end

    if settings["image/rotate"] then ic.corridor.enabled["#text"] = tostring(settings["image/rotate"].value) end
    if settings["image/upside_down"] then
        ic.ImageFlip.enabled["#text"] = tostring(settings["image/upside_down"].value)
        if settings["image/upside_down"].value then
            if ic.ImageFlip.ImageFlipStyle ~= nil then
                ic.ImageFlip.ImageFlipStyle["#text"] = "CENTER"
            else
                xml.dom.addElement(ic.ImageFlip, "ImageFlipStyle", "CENTER")
            end
        end
    end
    cs = xml.write(cs, {pretty=false})

    local response, _ = xml_request(self.imageUrl, "PUT", cs)
    if response == nil or response.ResponseStatus.statusString["#text"] ~= "OK" then return false end
    return true
end

function setup:set_cusom_osd(custom_id) -- custom id 1..5 (hik number)
    local current, _ = xml_request(self.osdUrl)
    if not table.is_subtable_seq(current, "VideoOverlay", "TextOverlayList") then return false end
    current = current.VideoOverlay.TextOverlayList -- could be empty dict without .TextOverlay

    local request = "<TextOverlayList>"
    for i = 1, 5 do
        local prefix = "osd/custom/" .. tostring(i - 1) .. "/"
        if settings[prefix .. "text"] then -- and (not custom_id or custom_id == i)
            local enabled = tostring( not string.is_empty(settings[prefix .. "text"].value) )
            local x, y = tostring(settings[prefix .. "pos"].value[1] or 0), tostring(settings[prefix .. "pos"].value[2] or 0)
            local text = tostring(settings[prefix .. "text"].value or "")
            local cur_overlay = "<TextOverlay><id>" .. tostring(i) .. "</id><enabled>" .. enabled .. "</enabled><positionX>" .. x .. "</positionX><positionY>" .. y .. "</positionY><displayText>" .. xml.escape(text) .. "</displayText></TextOverlay>"
            if enabled then request = request .. cur_overlay end

            -- NB! We use two setup methods for guarantee setup of OSD
            if not table.is_subtable_seq(current, "TextOverlay") or not current.TextOverlay[i] then
                local response, _ = xml_request(self.osdUrl .. "/text", "POST", cur_overlay)
                -- if response == nil or response.ResponseStatus.statusString["#text"] ~= "OK" then return false end
            else
                local old = current.TextOverlay[i]
                if string.is_empty(settings[prefix .. "text"].value) then
                    local response, _ = xml_request(self.osdUrl .. "/text/" .. tostring(i), "DELETE")
                    -- if response == nil or response.ResponseStatus.statusString["#text"] ~= "OK" then return false end
                elseif settings[prefix .. "text"].value ~= old.displayText["#text"]
                    or settings[prefix .. "pos"].value[1] ~= tonumber(old.positionX["#text"])
                    or settings[prefix .. "pos"].value[2] ~= tonumber(old.positionY["#text"]) then
                    local response, _ = xml_request(self.osdUrl .. "/text/" .. tostring(i), "PUT", cur_overlay)
                    -- if response == nil or response.ResponseStatus.statusString["#text"] ~= "OK" then return false end
                end
            end
        end
    end
    request = request .. "</TextOverlayList>"

    local response, _ = xml_request(self.osdUrl .. "/text", "PUT", request)
    if response == nil or response.ResponseStatus.statusString["#text"] ~= "OK" then return false end
    return true
end

function setup:get_channel_name()
    local chKey = (self.channelUrl:find("Proxy")) and "InputProxyChannel" or "VideoInputChannel"
    local channel, _ = xml_request(self.channelUrl)
    if settings["osd/name/text"] and table.is_subtable_seq(channel, chKey, "name") then
        settings["osd/name/text"].value = channel[chKey].name["#text"]
    else
        settings:remove("osd/name/text")
    end
end

function setup:set_channel_name()
    local chKey = (self.channelUrl:find("Proxy")) and "InputProxyChannel" or "VideoInputChannel"
    local request, _ = xml_request(self.channelUrl)
    if settings["osd/name/text"] and table.is_subtable_seq(request, chKey, "name") then
        request[chKey].name["#text"] = settings["osd/name/text"].value
    else
        return false
    end
    request = xml.write(request, {pretty=false})

    local response, _ = xml_request(self.channelUrl, "PUT", request)
    if response == nil or response.ResponseStatus.statusString["#text"] ~= "OK" then return false end
    return true
end

function setup:test_pir_available()
    if self.pir_available ~= nil then
        return self.pir_available
    end

    local response, _ = xml_request("/ISAPI/Event/triggers")
    if not table.is_subtable_seq(response, "EventNotification", "EventTriggerList", "EventTrigger") then
        settings:remove("pir_detection/enabled")
        return false
    end
    local events_array = response.EventNotification.EventTriggerList.EventTrigger

    self.pir_available = false
    for i = 1, #events_array do
        if events_array[i].id["#text"] == "PIR" then self.pir_available = true end
    end

    if not self.pir_available then
        settings:remove("pir_detection/enabled")
    end

    return self.pir_available
end

function setup:get_pir_settings()
    if not setup:test_pir_available() then return end

    local response, _ = xml_request(self.pirUrl)
    if response == nil or response.PIRAlarm == nil then
        settings:remove("pir_detection/enabled")
        return
    end

    settings["pir_detection/enabled"].value = (response.PIRAlarm.enabled["#text"] == "true")
end

function setup:set_pir_settings()
    if not setup:test_pir_available() then return false end

    local request = "<PIRAlarm><enabled>" .. tostring(settings["pir_detection/enabled"].value) .. "</enabled></PIRAlarm>"
    local response, _ = xml_request(self.pirUrl, "PUT", request)

    if response == nil or response.ResponseStatus.statusString["#text"] ~= "OK" then return false end
    return true
end

md.rt = "grid"
md.hik_unused_width_bits = nil

function md:init(hik_md_response)
    if not settings["motion_detector/0/zones"] then return end
    local md_meta = settings["motion_detector/0/zones"].meta
    local columns, rows = md_meta.range[1], md_meta.range[2]
    if not table.is_subtable_seq(hik_md_response, "MotionDetection", "regionType") then
        md.rt = md_meta._rt
    else
        md.rt = hik_md_response.MotionDetection.regionType["#text"]
        if md.rt == "grid" then
            local gs = hik_md_response.MotionDetection.Grid
            if type(gs) == "table" then rows, columns = tonumber(gs.rowGranularity["#text"]), tonumber(gs.columnGranularity["#text"]) end
        else  -- rt == "region"
            local rs = hik_md_response.MotionDetection.Region -- 9999, 9999 by default
            if type(rs) == "table" then columns, rows = tonumber(rs.normalizedScreenWidth["#text"]), tonumber(rs.normalizedScreenHeight["#text"]) end
        end
        md_meta.range = {columns, rows}
        md_meta._rt = md.rt
    end

    if md.rt == "grid" then
        local bytes_for_row = math.ceil(columns / 8)
        md.real_width = 8 * bytes_for_row
        md.hik_unused_width_bits = 8 * bytes_for_row - columns
    end

    md.raw_width, md.raw_height = columns, rows
    md.width_k, md.height_k = md.raw_width / md.ivideon_width, md.raw_height / md.ivideon_height
end

function md:hik_to_bingrid(hik_hex_str) -- return 01010101011101000 string with md.raw_width * md.raw_height chars
    local result = ""
    local hik_bin_str = self:hex2bin(hik_hex_str)
    for y = 1, md.raw_height do result = result .. hik_bin_str:sub((y - 1) * md.real_width + 1, y * md.real_width - md.hik_unused_width_bits) end

    return result
end

function md:bingrid_to_hik(bingrid)
    local result = ""
    for y = 1, md.raw_height do result = result .. md:bin2hex(bingrid:sub((y - 1) * md.raw_width + 1, y * md.raw_width) .. string.rep("0", md.hik_unused_width_bits)) end
    return result
end

function md:polyline_to_bingrid(hik_region_xml)
    local x_min, y_min, x_max, y_max = 1, md.raw_height-1, md.raw_width-1, 1
    if table.is_subtable_seq(hik_region_xml, "RegionCoordinatesList", "RegionCoordinates") then -- any polyline
        for _, v in pairs(hik_region_xml.RegionCoordinatesList.RegionCoordinates) do
            local x, y = tonumber(v.positionX["#text"]), tonumber(v.positionY["#text"])
            if x < x_min then x_min = x end
            if y < y_min then y_min = y end
            if x_max < x then x_max = x end
            if y_max < y then y_max = y end
        end
    end

    local x, width = x_min, x_max - x_min
    local y, height = md.raw_height - y_max, y_max - y_min
    return md:win_to_bingrid(x, y, width, height)
end

function setup:get_motion_detector()
    local response, _ = xml_request(self.mdUrl)
    md:init(response)

    local md_caps, _ = xml_request(self.mdUrl .. "/capabilities") -- 5.6.11+
    if table.is_subtable_seq(md_caps, "MotionDetection", "MotionDetectionLayout") then
        md_caps = md_caps.MotionDetection
    else
        md_caps, _ = xml_request(self.mdUrl .. "/layout/capabilities") -- up to 5.6.11
    end

    if table.is_subtable_seq(md_caps, "MotionDetectionLayout", "sensitivityLevel") then
        local senses = md_caps.MotionDetectionLayout.sensitivityLevel["#attr"]
        local r = {0, 100, 3}
        if senses.min then r[1] = tonumber(senses.min) end
        if senses.max then r[2] = tonumber(senses.max) end
        if senses.step then r[3] = tonumber(senses.step) end
        settings["motion_detector/0/sensitivity"].meta.range = r
    end

    if not settings["motion_detector/0/zones"] or not table.is_subtable_seq(response, "MotionDetection", "enabled") then
        settings:remove("motion_detector/enabled")
        settings:remove("motion_detector/0/sensitivity")
        settings:remove("motion_detector/0/zones")
        return false
    end
    settings["motion_detector/enabled"].value = (response.MotionDetection.enabled["#text"] == "true")
        if not table.is_subtable_seq(response, "MotionDetection", "MotionDetectionLayout", "layout") then
        settings:remove("motion_detector/0/zones")
        return false
    end

    local layout = response.MotionDetection.MotionDetectionLayout
    settings["motion_detector/0/sensitivity"].value = tonumber(layout.sensitivityLevel["#text"])
    layout = layout.layout

    if md.rt == "grid" then
        if not table.is_subtable_seq(layout, "gridMap") then return true end -- leave empty zone
        settings["motion_detector/0/zones"].value = md:hik_to_bingrid(layout.gridMap["#text"])
    else -- "region" with 4-point coordinates
        if not table.is_subtable_seq(layout, "RegionList", "Region") then return true end
        settings["motion_detector/0/zones"].value = md:polyline_to_bingrid(layout.RegionList.Region)
    end

    if string.is_empty(settings["motion_detector/0/zones"].value) then
        settings["motion_detector/0/zones"].value = string.rep('0', settings["motion_detector/0/zones"].meta.range[1] * settings["motion_detector/0/zones"].meta.range[2])
    end
end

function setup:set_motion_detector()
    if (settings["motion_detector/0/zones"] == nil or settings["motion_detector/0/zones"].value == nil) then return true end
    local request, _ = xml_request(self.mdUrl)
    md:init(request)
    local enabled = settings["motion_detector/enabled"].value

    if not table.is_subtable_seq(request, "MotionDetection", "MotionDetectionLayout", "sensitivityLevel") then return false end
    request.MotionDetection.enabled["#text"] = tostring(enabled)
    request.MotionDetection.MotionDetectionLayout.sensitivityLevel["#text"] = tostring(settings["motion_detector/0/sensitivity"].value)
    local layout = request.MotionDetection.MotionDetectionLayout.layout

    if md.rt == "grid" then
        if not table.is_subtable_seq(layout, "gridMap") then return false end
        layout.gridMap["#text"] = md:bingrid_to_hik(settings["motion_detector/0/zones"].value)
    else
        local x, y, width, height = md:extract_window_from_bingrid(settings["motion_detector/0/zone"].value)
        local x_min, y_max = math.max(1, x), math.min(md.raw_height, md.raw_height - y)
        local x_max, y_min = math.min(md.raw_width, x_min + width), math.max(1, y_max - height)
        local points = {{x_min, y_max}, {x_min, y_min}, {x_max, y_min}, {x_max, y_max}}

        if not xml.dom.testOrAddElements(layout, "RegionList", "Region", "RegionCoordinatesList", "RegionCoordinates") then return false end
        local rc = layout.RegionList.Region.RegionCoordinatesList.RegionCoordinates
        for i = 1, 4 do
            if not rc[i] then
                xml.dom.addElement(layout.RegionList.Region.RegionCoordinatesList, "RegionCoordinates")
                xml.dom.addElement(rc[i], "positionX")
                xml.dom.addElement(rc[i], "positionY")
            end
            rc[i].positionX["#text"] = points[i][1]
            rc[i].positionY["#text"] = points[i][2]
        end
    end

    local response, _ = xml_request(self.mdUrl, "PUT", xml.write(request))
    if response == nil or response.ResponseStatus.statusString["#text"] ~= "OK" then return false end
    return true
end

function setup:get_thermo_capability()

    if self.thermo_available ~= nil then
        return self.thermo_available
    end
    local caps, _ = xml_request(self.thermoUrl .. "/capabilities")
    if table.is_subtable_seq(caps, "ThermalCap", "isSupportRealTimethermometry") and (caps.ThermalCap.isSupportRealTimethermometry["#text"] == "true") then
        self.thermo_available = true
        return true
    end

    return false
end

function setup:get_anpr_capability()

    if self.anpr_available ~= nil then
        return self.anpr_available
    end
    local caps, _ = xml_request(self.anprUrl .. "/capability")
    if table.is_subtable_seq(caps, "ITCCap", "isSupportlicencePlateAuditData") and (caps.ITCCap.isSupportlicencePlateAuditData["#text"] == "true") then
        self.anpr_available = true
        return true
    end

    return false
end

function setup:get_acs_capability()

    if self.acs_available ~= nil then
        return self.acs_available
    end
    local caps, _ = xml_request(self.acsUrl .. "/capabilities")
    if table.is_subtable_seq(caps, "AccessControl", "isSupportAcsEvent") and (caps.AccessControl.isSupportAcsEvent["#text"] == "true") then
        self.acs_available = true
        return true
    end

    return false
end

function setup:get_acs_user_capability()


    if self.acs_user_available ~= nil then
        return self.acs_info
    end

    local acs_info = json.object()

    self.acs_user_available = false

    local resultUserCap, _ = http_request(self.acsUserUrl .. "/capabilities?format=json", "GET")
    local caps = json.parseStr(resultUserCap)
    if caps == json.NIL then
        acs_info.max_person_limit = 0
        acs_info.max_photos_per_person_limit = 0
        acs_info.max_cards_per_person_limit = 0
        acs_info.max_fp_per_person_limit = 0
        acs_info.user_verify_mode = 0

        self.acs_info = acs_info
        return self.acs_info
    end


    if table.is_subtable_seq(caps, "UserInfo") then
        acs_info.max_person_limit = caps.UserInfo.maxRecordNum
    end

    if ( acs_info.max_person_limit == 0 ) then
        self.acs_user_available = false
        self.acs_info = acs_info
        return self.acs_info
    end

    if table.is_subtable_seq(caps, "UserInfo", "userVerifyMode") then
        acs_info.user_verify_mode = caps.UserInfo.userVerifyMode["@opt"]
    end

    local resultCardCap, _ = http_request(self.acsCardUrl .. "/capabilities?format=json", "GET")
    local card_caps = json.parseStr(resultCardCap)

    if table.is_subtable_seq(card_caps, "CardInfo") then
        acs_info.max_cards_per_person_limit = tonumber(card_caps.CardInfo.maxRecordNum) / acs_info.max_person_limit
    end

    local resultFaceCap, _ = http_request(self.acsFaceLibUrl .. "/capabilities?format=json", "GET")
    local facelib_caps = json.parseStr(resultFaceCap)

    if facelib_caps ~= json.NIL then
        acs_info.max_photos_per_person_limit = tonumber(facelib_caps.FDRecordDataMaxNum) / acs_info.max_person_limit
    end

    local resultFingerPrintCap, _ = http_request(self.acsFingerprintUrl .. "/capabilities?format=json", "GET")
    local fingerprint_caps = json.parseStr(resultFingerPrintCap)

    if table.is_subtable_seq(fingerprint_caps, "FingerPrintCfg") then
        acs_info.max_fp_per_person_limit = 1
    end

    self.acs_user_available = true
    self.acs_info = acs_info
    return self.acs_info

end

-- Get fixed alarm input ID. There's an issue on the old devices: requests like "ISAPI/System/IO/inputs"
-- return integer port IDs with custom prefixes (for example, "<id>I-1</id>" for port 1), while event
-- notifications in the isapi and binary protocols contain only pure integer port IDs. So here we remove
-- the prefix and leave only the last integer part of the ID
local function get_fixed_alarm_input_id(id)
    return string.match(id or "", "^.-(%d+)$")
end

function parse_alarm_inputs_info(ports, result, withActive, fullList, only_port)

    for _, p in ipairs(ports) do
        local portId = get_fixed_alarm_input_id(p.id["#text"])
        if (fullList or (only_port == portId)) then
            local port = json.object()
            if (p.name ~= nil and p.name["#text"] ~= "") then
                port.name = p.name["#text"]
            end
            if withActive then
                if (p.enabled ~= nil and p.enabled["#text"] == "true") then
                    port.active = true
                else
                    port.active = false
                end
            end
            result[portId] = port
        end
    end

    return result
end

function request_alarm_inputs_info(withActive, fullList, only_port)
    local result
    local portResult = {}

    local response, code  = xml_request("/ISAPI/System/IO/inputs")

    if response and table.is_subtable_seq(response, "IOInputPortList", "IOInputPort") then
        local ports = response.IOInputPortList.IOInputPort
        if type(ports) == "table" then
            if #ports == 0 then
                -- We expected "IOInputPortList.IOInputPort" to be array but it's just a single element. So transform it to array
                ports = {ports}
            end

            portResult = parse_alarm_inputs_info(ports, portResult, withActive, fullList, only_port)
        end
    end

    -- Note: proxied IO inputs are not supported yet. If you want to add them then uncomment the next line
    -- local response2, code2  = xml_request("/ISAPI/ContentMgmt/IOProxy/inputs")

    if response2 and table.is_subtable_seq(response2, "IOProxyInputPortList", "IOProxyInputPort") then
        local ports = response2.IOProxyInputPortList.IOProxyInputPort
        if type(ports) == "table" then
            if #ports == 0 then
                -- We expected "IOInputPortList.IOInputPort" to be array but it's just a single element. So transform it to array
                ports = {ports}
            end

            portResult = parse_alarm_inputs_info(ports, portResult, withActive, fullList, only_port)
        end
    end

    if next(portResult) ~= nil then
        result = portResult
    end

    if (code == http_request.OK_CODE or code2 == http_request.OK_CODE) then
        code = http_request.OK_CODE
    end

    return result, code
end


function setup:get_alarm_inputs_capability()

    if self.digital_inputs_ports ~= nil then
        return  self.digital_inputs_ports
    end

    local result, code = request_alarm_inputs_info(false, true)

    if result ~= nil and next(result) ~= nil and code == http_request.OK_CODE then
        self.digital_inputs_ports = result
        return self.digital_inputs_ports
    end

end

function setup:test_twoway_audio_available()
    if self.twoway_audio_available ~= nil then
        return self.twoway_audio_available
    end

    local request, _ = xml_request("/ISAPI/System/Audio/capabilities") -- to test is audio device available
    if type(request) == "table" and not setup.isProxied and table.is_subtable_seq(request, "AudioCap", "audioInputNums") then
        local inputCounter = tonumber(request.AudioCap.audioInputNums["#text"])
        if inputCounter > 0 then
            self.twoway_audio_available = true
            return true
        end
    end

    -- twoway audio is unavailable
    self.twoway_audio_available = false
    settings:remove("microphone/sensitivity")
    settings:remove("microphone/denoise")
    return false
end

function setup:get_twoway_audio_settings()
    local caps = {}
    if setup:test_twoway_audio_available() then caps, _ = xml_request("/ISAPI/System/TwoWayAudio/channels/" .. setup.cameraId .. "/capabilities") end
    if not table.is_subtable_seq(caps, "TwoWayAudioChannel") then
        settings:remove("microphone/sensitivity")
        settings:remove("microphone/denoise")
        return
    end
    caps = caps.TwoWayAudioChannel
    -- <TwoWayAudioChannel><id>1</id>
    -- <enabled>false</enabled>
    -- <audioCompressionType opt="G.722.1,G.711ulaw,G.711alaw,MP2L2,G.726,PCM">G.711ulaw</audioCompressionType>
    -- <audioInputType opt="MicIn,LineIn">LineIn</audioInputType>
    -- <speakerVolume min="0" max="100">60</speakerVolume>
    -- <noisereduce opt="true,false">false</noisereduce></TwoWayAudioChannel>

    if not caps.speakerVolume or not caps.speakerVolume["#attr"].min then
        settings:remove("microphone/sensitivity")
    else
        settings["microphone/sensitivity"].value = tonumber(caps.speakerVolume["#text"])
        settings["microphone/sensitivity"].meta.range = {tonumber(caps.speakerVolume["#attr"].min), tonumber(caps.speakerVolume["#attr"].max)}
    end

    if not caps.noisereduce then
        settings:remove("microphone/denoise")
    else
        settings["microphone/denoise"].value = (caps.noisereduce["#text"] == "true")
        local variants = caps.noisereduce["#attr"].opt
        if type(variants) == "string" then variants = variants:split(",") else variants = {} end
        if #variants <= 1 then settings:remove("microphone/denoise") end -- can be forced off/on
    end
end

function setup:set_twoway_audio_settings()
    if not setup:test_twoway_audio_available() then return false end
    local url = "/ISAPI/System/TwoWayAudio/channels/" .. setup.cameraId
    local request, _ = xml_request(url)
    if not table.is_subtable_seq(request, "TwoWayAudioChannel", "speakerVolume") then return false end

    request.TwoWayAudioChannel.speakerVolume["#text"] = tostring(settings["microphone/sensitivity"].value)

    if settings["microphone/denoise"] and request.TwoWayAudioChannel.noisereduce then
        request.TwoWayAudioChannel.noisereduce["#text"] = tostring(settings["microphone/denoise"].value)
    end

    request = xml.write(request, {pretty=false})
    local response, _ = xml_request(url, "PUT", request)
    if response == nil or response.ResponseStatus.statusString["#text"] ~= "OK" then return false end
    return true
end

function settings:update()
    self:update_base_impl()

    -- Work-around for 2532 issue and secondary stream's settings break:
    if settings["streams/1/resolution"] ~= nil and settings["streams/1/resolution"].value == nil then
        safe_call(setup.set_stream_settings, setup, 0, true)
        safe_call(setup.load_streams_settings, setup)
    end
end

function setup_settings_handlers() -- global
    table.merge(settings.traits, {
        ["general/online_mode"] = {get = setup.get_online_mode_state, set = setup.save_online_mode},
        ["general/timezone"] = {get = setup.get_timezone, set = setup.set_timezone},
        ["night_vision/mode"] = {get = setup.get_irled_state, set = setup.save_irled},
        ["microphone/enabled"] = {get = setup.load_streams_settings, set = setup.set_streams_settings},
        ["microphone/sensitivity"] = {get = setup.get_twoway_audio_settings, set = setup.set_twoway_audio_settings},
        ["microphone/denoise"] = {get = setup.get_twoway_audio_settings, set = setup.set_twoway_audio_settings},
        ["pir_detection/enabled"] = {get = setup.get_pir_settings, set = setup.set_pir_settings},
        ["osd/name/text"] = {get = setup.get_channel_name, set = setup.set_channel_name},
    })

    settings("microphone/enabled").is_changed_hint = function(value)
        -- WORK-AROUND It's common setting that can be set on both streams, check it out
        -- Note: see also VS-2700 if it is bug
        return (value ~= (settings.sound_enabled[1] or settings.sound_enabled[2]))
    end

    for _, v in ipairs({"enabled", "0/zones", "0/sensitivity"}) do
        settings("motion_detector/" .. v).get = setup.get_motion_detector
        settings("motion_detector/" .. v).set = setup.set_motion_detector
    end
    for _, stream in ipairs({"0", "1"}) do for _, v in ipairs({"resolution", "bitrate", "fps", "codec", "acodec", "i_frame_interval", "variable_bitrate"}) do
        settings("streams/" .. stream .. "/" .. v).get = setup.load_streams_settings
        settings("streams/" .. stream .. "/" .. v).set = function(setup) return setup.set_stream_settings(setup, tonumber(stream)) end
    end end
    for _, v in ipairs({"transparent", "flashing", "datetime/enabled", "datetime/display_week", "datetime/24h", "datetime/format", "datetime/pos", "name/enabled", "name/pos"}) do
        settings("osd/" .. v).get = setup.get_osd_settings
        settings("osd/" .. v).set = setup.set_osd_settings
    end

    for i = 1, 5 do for _, v in ipairs({"text", "pos"}) do
        settings("osd/custom/" .. tostring(i - 1) .. "/" .. v).get = setup.get_osd_settings
        settings("osd/custom/" .. tostring(i - 1) .. "/" .. v).set = function(setup) return setup.set_cusom_osd(setup, i) end
    end end

    for _, v in ipairs({"brightness", "contrast", "saturation", "hue", "upside_down", "rotate", "wb", "shutter_time", "gain"}) do
        settings("image/" .. v).get = setup.get_image_settings
        settings("image/" .. v).set = setup.set_image_settings
    end
    settings("general/power_freq").get = setup.get_image_settings
    settings("general/power_freq").set = setup.set_image_settings

    settings.traits.access_funcs_owner_table = setup
end

function racm_substream_count(channel) -- 1-3 for NVR, 4 for analog devices
    local substream_count = 4
    local status, _ = xml_request( string.format("/ISAPI/ContentMgmt/InputProxy/channels/%s/status", channel) )
    if type(status) == "table" and type(status.InputProxyChannelStatus) == "table" and type(status.InputProxyChannelStatus.streamingProxyChannelIdList) == "table" then
        substream_count = 1 -- default for offline proxied cams
        local substreams = status.InputProxyChannelStatus.streamingProxyChannelIdList.streamingProxyChannelId
        if type(substreams) == "table" and substreams["#name"] == nil then
            substream_count = table.getn(substreams)
        end
    end

    return substream_count
end

function setup_api_request_params(cdi_url)
    http_request.base_url = (cdi_url or os.getenv("BASE_URL"))
    if not http_request.base_url then error("Couldn't get api url") end

    http_request.timeout_ms = 12000
    http_request.max_retries_count = 3
    http_request.retry_delay_ms = 1000
end

local function is_ptz_available(channels_xml, channel)
    if not table.is_subtable_seq(channels_xml, "PTZChannelList", "PTZChannel") then return false end
    for _, v in pairs(channels_xml.PTZChannelList.PTZChannel) do
        if table.is_subtable_seq(v, "enabled") and v.enabled["#text"] == "true" then
            if channel ~= nil then
                if type(v.id) == "table" and channel == v.id["#text"] then return true end
            else -- IPC: PTZ enabled also for FishEye-cameras, to filter out it:
                local tiltOrZoomSupported = (type(v.tiltSupport) == "table" and v.tiltSupport["#text"] == "true")
                                         or (type(v.zoomSupport) == "table" and v.zoomSupport["#text"] == "true")
                if v.enabled["#text"] == "true" and tiltOrZoomSupported then return true end
            end
        end
    end

    return false
end

local function is_ptz_proxy_available(channel)
    -- NB Cameras must be connected and ready to perform this action (dummy)
    local response, _ = xml_request("/ISAPI/ContentMgmt/PTZCtrlProxy/channels/" .. channel .. "/continuous", "PUT", "<PTZData><pan>0</pan><tilt>0</tilt><zoom>0</zoom></PTZData>")
    if table.is_subtable_seq(response, "ResponseStatus", "statusString") and response.ResponseStatus.statusString["#text"] == "OK" then return true else return false end
end

function detect_streams(config)
    local function request_streams(search_type, rtsp_base, template, found_cams, archive_suffix)
        local func_setup = {
            intern = {uri = "/ISAPI/System/Video/inputs/channels", items_name = "VideoInputChannel", ptz_uri = "/ISAPI/PTZCtrl/channels"},
            extern = {uri = "/ISAPI/ContentMgmt/InputProxy/channels", items_name = "InputProxyChannel", ptz_uri = "/ISAPI/ContentMgmt/PTZCtrlProxy/channels"}
        }

        local streams, _ = xml_request(func_setup[search_type].uri)
        if not table.is_subtable_seq(streams, func_setup[search_type].items_name .. "List", func_setup[search_type].items_name) then return archive_suffix end
        local ptzs, _ = xml_request(func_setup[search_type].ptz_uri)

        for _, v in pairs( streams[func_setup[search_type].items_name .. "List"][func_setup[search_type].items_name] ) do
            if type(v.id) == "table" and type(v.name) == "table" then
                local new_cam = table.copy(template)
                new_cam.id = template.id
                new_cam.name = v.name["#text"]

                local channel = v.id["#text"]
                new_cam.urlHigh = rtsp_base .. channel .. "01"

                if search_type == "extern" and racm_substream_count(channel) < 2 then new_cam.urlMedium = nil end
                if not string.is_empty(new_cam.urlMedium) then new_cam.urlMedium = rtsp_base .. channel .. "02" end

                new_cam.usePtz = false
                -- NB! PTZ auto-detection doesn't works on DVR channels (always enabled) is_ptz_available(ptzs, channel)
                if ( search_type == "extern" and is_ptz_proxy_available(channel) ) then
                    local tmp_url = iv.parse_url(new_cam.ptzUrl)
                    tmp_url.query = "channel=" .. channel
                    if search_type == "extern" then tmp_url.scheme = "hik-ptz-proxy" end
                    new_cam.ptzUrl = iv.build_url(tmp_url)
                    new_cam.usePtz = true
                end

                archive_suffix = string.format("%s%d=%s;", archive_suffix, (template.id << 16), channel)

                found_cams[#found_cams + 1] = new_cam
                template.id = template.id + 1
            end
        end
        return archive_suffix
    end

    local archive_suffix = "?stream_map="
    local cam = config.cameras[1]
    if table.is_subtable_seq(cam, "streamHigh", "videoBitrateKbps") then cam.streamHigh.videoBitrateKbps = 1536 end
    if table.is_subtable_seq(cam, "streamMedium", "videoBitrateKbps") then cam.streamMedium.videoBitrateKbps = 256 end
    local found_cams = {}

    local rtsp_base = cam.urlHigh:sub(1, -cam.urlHigh:reverse():find("/"))
    local caps, _ = xml_request("/ISAPI/System/capabilities")
    if not table.is_subtable_seq(caps, "DeviceCap", "SysCap") then error("Failed to parse capabilities:", json.write(caps)) end
    caps = caps.DeviceCap -- raises "failed to index" exception in case or issues

    if table.is_subtable_seq(caps, "SysCap", "VideoCap", "videoInputPortNums") then
        if tonumber(caps.SysCap.VideoCap.videoInputPortNums["#text"]) == 0 then -- IPC
            local ptzCtrlResponse, _ = xml_request("/ISAPI/PTZCtrl/channels")
            if is_ptz_available(ptzCtrlResponse, nil) then cam.usePtz = true end
        else -- DVR
            archive_suffix = request_streams("intern", rtsp_base, cam, found_cams, archive_suffix)
        end
    end

    if table.is_subtable_seq(caps, "RacmCap", "inputProxyNums") then
        -- RacmCap struct has proxied flags for: isSupportXXX: Raid, ExtHdCfg, TransCode, IpcImport, IpcStreamType, IOInputProxy, SMARTTest, ActivateIpc, IPCPassword, Transfer, SmartSeaarch, POS
        archive_suffix = request_streams("extern", rtsp_base, cam, found_cams, archive_suffix)
    end

    if #found_cams > 0 then
        config.cameras = found_cams
        config.archive.path = config.archive.path:split("?")[1] .. archive_suffix
    end
end

function get_relay_outputs_info()
    local result
    local portResult = {}
    local response, code = xml_request("/ISAPI/System/IO/status")
    if not response or not response.IOPortStatusList then
        return result, code
    end

    local ports = response.IOPortStatusList.IOPortStatus
    if type(ports) ~= "table" then
        return
    end

    if #ports == 0 then
        -- We expected "IOPortStatusList.IOPortStatus" to be array but it's just a single element. So transform it to array
        ports = {ports}
    end

    for _, p in ipairs(ports) do
        if p.ioPortType["#text"] ~= "output" then
            goto next
        end
        portResult[p.ioPortID["#text"]] = {mode = "bistable"}
        ::next::
    end

    if next(portResult) ~= nil then
        result = portResult
    end

    return result, code
end

function setup:get_relay_outputs_capability()

    if self.relay_outputs_ports ~= nil then
        return self.relay_outputs_ports
    end

    local result, code = get_relay_outputs_info()

    if result ~= nil and next(result) ~= nil and code == http_request.OK_CODE then
        self.relay_outputs_ports = result
        return self.relay_outputs_ports
    end

end

function convert_user_verify_mode(in_verify_mode)
    local result_str=""
    case_action = {
       ["card"] = function ()  return "card" end,
       ["cardAndPw"] = function () return "card&pw"  end,
       ["cardOrFace"] = function () return "card|face"  end,
       ["cardOrFaceOrFp"] = function () return "card|face|fp"  end,
       ["cardOrFpOrPw"] = function () return "card|fp|pw"  end,
       ["cardOrPw"] = function () return "card|pw"  end,
       ["cardOrfaceOrPw"] = function () return "card|face|pw"  end,
       ["employeeNoAndFace"] = function () return nil  end,
       ["employeeNoAndFp"] = function () return nil  end,
       ["employeeNoAndFpAndPw"] = function () return nil  end,
       ["employeeNoAndPw"] = function () return nil  end,
       ["face"] = function () return "face"  end,
       ["faceAndCard"] = function () return "face&card"  end,
       ["faceAndFp"] = function () return "face&fp"  end,
       ["faceAndFpAndCard"] = function () return "face&fp&card"  end,
       ["faceAndPw"] = function () return "face&pw"  end,
       ["faceOrPw"] = function () return "face|pw"  end,
       ["faceAndPwAndFp"] = function () return "face&pw&fp"  end,
       ["faceOrFpOrCardOrPw"] = function () return "face|fp|card|pw"  end,
       ["faceOrfaceAndCard"] = function () return "face|face&card"  end,
       ["fp"] = function () return "fp"  end,
       ["fpAndCard"] = function () return "fp&card"  end,
       ["fpAndCardAndPw"] = function () return "fp&card&pw"  end,
       ["fpAndPw"] = function () return "fp&pw"  end,
       ["fpOrCard"] = function () return "fp|card"  end,
       ["fpOrPw"] = function () return "fp|pw"  end,
       ["fpOrface"] = function () return "fp|face"  end,
    }
    for line in in_verify_mode:gsub("%f[,]%,%f[^,]", "\0"):gmatch"%Z+" do
        if (case_action[line] ~= nil) then
            local convert_res = case_action[line]()
            local sep=","
            if result_str == "" then sep = "" end
            if convert_res ~= nil then
                result_str=result_str..sep..convert_res
            end
        end
    end

    return result_str
end

function get_info_features()
    local features = json.object()
    local result = false
    local thermoResult = setup:get_thermo_capability() -- If there is thermovision settings we'll append thermovision feature
    if thermoResult then
        features["thermal"] = {
            ["provider"] = "hikvision",
            ["version"] = "1.0"
        }
    end

    local anprResult = setup:get_anpr_capability() -- If there is ITC License plate settings we'll append ANPR feature
    if anprResult then
        features.anpr = {
            protocol = "hikvision",
            number_plate_max_limit = 10000,
            functions = {
                "anpr.list.put"
            }
        }
    end

    local acsResult = setup:get_acs_capability() -- If there is Access Control Event
    if acsResult then
        features.acs = {
            provider = "hikvision",
            version = "1.0",
            functions = {
                "acs.person.put",
                "acs.person.delete",
                "acs.person.delete_all",
            }
        }

        local acs_info = setup:get_acs_user_capability()

        features.acs.max_person_limit = acs_info.max_person_limit
        features.acs.max_photos_per_person_limit = acs_info.max_photos_per_person_limit
        features.acs.max_cards_per_person_limit = acs_info.max_cards_per_person_limit
        features.acs.max_fp_per_person_limit = acs_info.max_fp_per_person_limit
        features.acs.auth_methods = convert_user_verify_mode(acs_info.user_verify_mode)
        iv.log("Add Hikvision ACS feature "..json.write(features.acs))
    end

    local digitalInputsPorts = setup:get_alarm_inputs_capability() -- If there is alarm inputs settings we'll append digital_inputs feature
    if digitalInputsPorts ~= nil and next(digitalInputsPorts) ~= nil then
        features.digital_inputs = {
            ports = digitalInputsPorts,
            functions = {
                "device.ctl.digital_input.status"
            }
        }
        iv.log("Add Hikvision Digital Input feature "..json.write(features.digital_inputs))
    end

    local relayOutputsPorts = setup:get_relay_outputs_capability() -- If there is alarm inputs settings we'll append relay_outputs feature
    if relayOutputsPorts ~= nil and next(relayOutputsPorts) ~= nil then
        features.relay_outputs = {
            ports = relayOutputsPorts,
            functions = {
                "device.ctl.relay_output.status"
            }
        }
        iv.log("Add Hikvision Relay Outputs feature "..json.write(features.relay_outputs))
    end

    return features
end

local function get_devinfo_field(response, field_name)
    if type(response) == "table" and type(response.DeviceInfo) == "table" and type(response.DeviceInfo[field_name]) == "table" then
        return tostring(response.DeviceInfo[field_name]["#text"])
    end
end

function get_info()
    local response, _ = xml_request("/ISAPI/System/deviceInfo")
    if not response then -- any XML API error including parse errors
        return nil
    end

    local device_info = json.object()
    device_info.serial_number = get_devinfo_field(response, "serialNumber")
    device_info.vendor = settings.vendor_name
    device_info.model = get_devinfo_field(response, "model")
    device_info.fw_version = get_devinfo_field(response, "firmwareVersion")
    device_info.hardware_id = get_devinfo_field(response, "hardwareVersion")
    device_info.mac_address = get_devinfo_field(response, "macAddress")
    device_info.serial_number_format = "hikvision"
    device_info.features = get_info_features()

    return device_info
end

function probe(cmd)
    http_request.max_retries_count = 1

    local reply = cmd
    if not cmd.args.inspection_protocol then
        cmd.args.inspection_protocol = cmd.args.cdi_scheme or "isapi"
    end

    local device_info, status = xml_request("/ISAPI/System/deviceInfo")
    if status ~= nil and status ~= 200 then
        reply.args = { status = status,  reason = "failed to request device info" }
        return reply
    end

    local result = cmd.args
    result.manufacturer = "Hikvision"
    result.model = get_devinfo_field(device_info, "model")
    result.description = get_devinfo_field(device_info, "deviceDescription")
    result.presentation_url = (cmd.args.use_https and "https://" or "http://") .. cmd.args.host .. ":" .. tostring(cmd.args.port)
    result.sn = get_devinfo_field(device_info, "serialNumber")
    result.mac = get_devinfo_field(device_info, "macAddress")
    result.channels = { }

    local accesses, status = xml_request("/ISAPI/Security/adminAccesses")
    if status ~= nil and status ~= 200 then
        reply.args = { status = status,  reason = "failed to request access protocol list" }
        return reply
    end

    local rtsp_port = "554"
    if table.is_subtable_seq(accesses, "AdminAccessProtocolList", "AdminAccessProtocol") then
        for _, v in pairs(accesses.AdminAccessProtocolList.AdminAccessProtocol) do
            if table.is_subtable_seq(v, "protocol") and v.protocol["#text"] == "RTSP" then
                rtsp_port = v.portNo["#text"]
                break
            end
        end
    end

    local storages, status = xml_request("/ISAPI/ContentMgmt/Storage")
    if status ~= nil and status ~= 200 then
        reply.args = { status = status,  reason = "failed to request storage info" }
        return reply
    end

    local storage_count = 0
    for _, name in ipairs({ "hdd", "nas" })  do
        if table.is_subtable_seq(storages, "storage", name .. "List", name) then
            for _, v in pairs(storages.storage[name .. "List"][name]) do
                if table.is_subtable_seq(v, "status") and v.status["#text"] == "ok" then
                    storage_count = storage_count + 1
                end
            end
        end
    end

    local funcs = {
        analog = {
            uri = "/ISAPI/System/Video/inputs/channels",
            item_name = "VideoInputChannel",
            ptz_uri = "/ISAPI/PTZCtrl/channels"
        },
        proxy = {
            uri = "/ISAPI/ContentMgmt/InputProxy/channels",
            item_name = "InputProxyChannel",
            ptz_uri = "/ISAPI/ContentMgmt/PTZCtrlProxy/channels"
        }
    }

    for k, p in pairs(funcs) do
        local input_channels, status = xml_request(p.uri)
        if status ~= nil and status ~= 200 then
            if status == 403 then goto continue end -- for IPC
            reply.args = { status = status,  reason = "failed to request list of input channels" }
            return reply
        end

        if not table.is_subtable_seq(input_channels, p.item_name .. "List", p.item_name) then
            goto continue
        end

        local ptz_channels, status = xml_request(p.ptz_uri)
        if status ~= nil and status ~= 200 then
            reply.args = { status = status,  reason = "failed to request list of ptz channels" }
            return reply
        end

        for _, v in pairs(input_channels[p.item_name .. "List"][p.item_name]) do
            if type(v.id) ~= "table" then
                goto next
            end

            -- N = id * 100 + T;
            -- T: 1-main streaming, 2-sub streaming
            local ch   = v.id["#text"]
            local id   = tonumber(ch)
            local main = id * 100 + 1
            local sub  = id * 100 + 2

            local channel = { }
            channel.hw_id = tostring(id)
            channel.video = {
                primary = http_request:copy_url_with{scheme = "rtsp", port = rtsp_port, path = "/Streaming/channels/" .. tostring(main)},
                secondary = http_request:copy_url_with{scheme = "rtsp", port = rtsp_port, path = "/Streaming/channels/" .. tostring(sub)},
                preview = http_request:copy_url_with{path = "/ISAPI/Streaming/channels/" .. tostring(main) .. "/picture?snapShotImageType=JPEG"},
                rtsp_transport = "tcp"
            }

            channel.events = "isapi?port=" .. tostring(cmd.args.port)
            channel.config = cmd.args.cdi_scheme .. "?port=" .. tostring(cmd.args.port)
            if k == "analog" then
                if is_ptz_available(ptz_channels, ch) then
                    channel.ptz = http_request:copy_url_with{scheme = "hik-ptz", query = "channel=" .. ch}
                end
            elseif is_ptz_proxy_available(ch) then
                channel.ptz = http_request:copy_url_with{scheme = "hik-ptz-proxy", query = "channel=" .. ch}
            end

            if storage_count > 0 then
                channel.storage = http_request:copy_url_with{scheme = "isapi", query = (cmd.args.use_https and "https=true" or "")}
            end

            result.channels[#result.channels + 1] = channel
            ::next::
        end

        ::continue::
    end

    reply.args = { status = 200,  result = result }
    return reply
end

function test_config(method, args) -- REMOVEME
    return cloud_config_update(args)
end

caps.requests.get_info = {
    useCache = true,
    level = "mod"
}
