require("cloud_config_template")

settings:add({
    { name = "night_vision/adaptive/enabled", meta = {type = "bool"} },
    { name = "microphone/sensitivity", meta = {type = "int", range = {0, 255}} },
    { name = "microphone/denoise", meta = {type = "bool"} },
    { name = "microphone/source", meta = {type = "choice"} },
    { name = "motion_detector/0/sensitivity", meta = {type = "int", range = {0, 100, 5}} },
    { name = "motion_detector/0/zones", meta = {type = "grid", range = {}} },
    { name = "osd/transparent", meta = {type = "bool"} },
    { name = "osd/datetime/enabled", meta = {type = "bool"} },
    { name = "osd/datetime/format", meta = {type = "choice", range = {} } },
    { name = "osd/datetime/pos", meta = {type = "position", range = {} } },
    { name = "osd/name/enabled", meta = {type = "bool"} },
    { name = "osd/name/text", meta = {type = "string", range = {0, 32}} },
    { name = "osd/name/pos", meta = {type = "position", range = {} } },
    { name = "image/brightness", meta = {type = "int", range = {0, 255}} },
    { name = "image/contrast", meta = {type = "int", range = {0, 255}} },
    { name = "image/saturation", meta = {type = "int", range = {0, 255}} },
    { name = "image/upside_down", meta = {type = "bool"} },
    { name = "image/rotate", meta = {type = "bool"} }})

for i = 0, 5 do settings:add({
    { name = "osd/custom/" .. tostring(i) .. "/text", meta = {type = "string", range = {0, 32}} },
    { name = "osd/custom/" .. tostring(i) .. "/pos", meta = {type = "position", range = {} } },
}) end

settings:remove_base("general/power_freq")

settings.traits = {
    ["streams/0/codec"] = {
        depends = { "streams/0/bitrate", "streams/0/fps" },
        restart_requered = true
    },
    ["streams/0/acodec"] = {
        restart_requered = true
    },
    ["streams/1/codec"] = {
        depends = { "streams/1/bitrate", "streams/1/fps" },
        restart_requered = true
    },
    ["streams/2/codec"] = {
        depends = { "streams/2/bitrate", "streams/2/fps" },
        restart_requered = true
    },
    ["streams/0/resolution"] = {
        depends = { "streams/0/bitrate", "streams/0/fps", "streams/1/resolution", "streams/1/bitrate", "streams/1/fps", "streams/2/resolution", "streams/2/bitrate", "streams/2/fps" },
        restart_requered = true
    },
    ["streams/1/resolution"] = {
        depends = { "streams/1/bitrate", "streams/1/fps" },
        restart_requered = true
    },
    ["streams/2/resolution"] = {
        depends = { "streams/2/bitrate", "streams/2/fps" },
        restart_requered = true
    },
    ["streams/0/fps"] = {
        depends = {"streams/0/bitrate"}
    },
    ["streams/1/fps"] = {
        depends = {"streams/1/bitrate"}
    },
    ["streams/2/fps"] = {
        depends = {"streams/2/bitrate"}
    },
    ["general/timezone"] = {
        restart_requered = true
    },
    ["microphone/enabled"] = {
        restart_requered = true
    },
    ["osd/datetime/enabled"] = {
        restart_requered = true
    },
    ["osd/name/enabled"] = {
        restart_requered = true
    },
    ["image/upside_down"] = {
        restart_requered = true
    },
    ["image/rotate"] = {
        depends = {"image/upside_down"},
        restart_requered = true
    },
    ["microphone/enabled"] = {
        restart_requered = true
    },
    ["microphone/source"] = {
        restart_requered = true
    }
}

settings.timezones = { "-12:00", "-11:00", "-10:00", "-09:00", "-08:00", "-07:00", "-06:00",
                       "-05:00", "-04:30", "-04:00", "-03:30", "-03:00", "-02:00", "-01:00",
                       "-00:00", "+01:00", "+02:00", "+03:00", "+03:30", "+04:00", "+04:30",
                       "+05:00", "+05:30", "+05:45", "+06:00", "+06:30", "+07:00", "+08:00",
                       "+09:00", "+09:30", "+10:00", "+11:00", "+12:00", "+13:00" }

settings.codecs = { ["0"] = "MJPEG", ["1"] = "H.264", ["2"] = "H.265", ["3"] = "JPEG" }
settings.bitrates = { 64, 128, 256, 384, 512, 768, 1024, 1536, 2048, 3072, 4096, 5120, 6144,
                      8192, 10240, 12288, 14336, 16384, 20480, 24576, 30720, 40960 }
settings.fpses = { 1, 3, 5, 6, 8, 10, 12, 15, 20, 25, 30, 50, 60 }

settings.acodecs = { [0] = "MPEG1", [1] = "G.711A", [2] = "G.711U", [3] = "ADPCM", [4] = "OGG",
                     [5] = "Auto", [6] = "AAC-LC" }
settings.asources = { [0] = "LineIn", [1] = "Mic", [2] = "RS485" }
settings.afreqs = { [0] = "8 kHz", [1] = "16 kHz", [2] = "22 kHz", [3] = "24 kHz", [4] = "32 kHz",
                    [5] = "44 kHz", [6] = "48 kHz", [7] = "96 kHz" }

-- dd=Day; dddd=Day of the week; M=Month; y=Year
settings.date_format = { [0]   = "yyyy-MM-dd",
                         [1]   = "MM-dd-yyyy",
                         [2]   = "yyyyMMdd",
                         [3]   = "MMddyyyy",
                         [4]   = "yyyyMMdd X",
                         [5]   = "MMddyyyy X",
                         [100] = "dd/MM/yyyy",
                         [101] = "MM/dd/yyyy",
                         [102] = "dd MMMM, yyyy",
                         [103] = "MMMM dd, yyyy",
                         [104] = "dddd, dd MMMM, yyyy",
                         [105] = "dddd, MMMM dd, yyyy",
                         [106] = "yyyy/MM/dd",
                         [107] = "yyyy, MMMM dd",
                         [108] = "dddd, yy, MM dd" }

-- h/H=12/24 Hour; tt=A.M. or P.M.; mm=Minute; ss=Second
settings.time_format = { [0] = "HH:mm:ss",
                         [1] = "hh:mm:ss t.t.",
                         [2] = "hh:mm:ss tt",
                         [3] = "tt hh:mm:ss",
                         [4] = "hh:mm:ss",
                         [5] = "HH:mm:ss.xxx",
                         [6] = "hh:mm:ss.xxx tt" }

settings.device_type = { [0] = "IPC",
                         [1] = "NVR",
                         [2] = "DVR",
                         [3] = "VMS" }

--
-- HELPERS
--

if type(http_request.call_orig) ~= "function" then
    http_request.call_orig = http_request.__call
    http_request.__call = function(...)
        local raw, code = http_request.call_orig(...)
        if type(raw) == "string" then
            local response = json.parseStr(raw).Response
            if response and response.ResponseCode == 0 then
                return response.Data, code
            end
        end

        return nil, code
    end
end

function bool_to_number(value)
    return value and 1 or 0
end

function number_to_binary_string(num, bits)
    local str = ""
    for i = bits, 1, -1 do
        rest = math.fmod(num, 2)
        str = string.format("%d%s", rest, str)
        num = (num - rest) / 2
    end

    return str
end

function binary_string_to_number(str)
    str = string.reverse(str)

    local sum = 0
    for i = 1, #str do
        num = string.sub(str, i, i) == "1" and 1 or 0
        sum = sum + num  * math.floor(2 ^ (i - 1))
    end

    return sum
end

function unsigned(x)
    if x < 0 then return 256 + x
    else return x end
end

--
-- SETUP
--

local stream_params = { "codec", "resolution", "bitrate", "fps", "i_frame_interval", "variable_bitrate" }

setup = {
    channel = 0
}

function setup:set_camera(camera)
    local info = vs.get_camera_info(tonumber(camera))
    if type(info) ~= "table" then return end

    if log_config.tag == tostring(camera) and type(settings["general/online_mode"]) == "table" then return end -- already set
    log_config.tag = tostring(camera)

    if info.id then setup.channel = info.id end
    settings:load_from_storage(tonumber(camera))
end

function setup:get_online_mode()
    local urls = {
        "/LAPI/V1.0/Channels/" .. self.channel .. "/Alarm/MotionDetection/WeekPlan",
        "/LAPI/V1.0/Channels/" .. self.channel .. "/Alarm/AudioDetection/WeekPlan"
    }

    for _, url in ipairs(urls) do
        local plan = http_request(url)
        if plan and plan.Enabled == 1 then
            settings["general/online_mode"].value = true
            return
        end
    end

    settings["general/online_mode"].value = false
end

function setup:set_online_mode()
    if not settings["general/online_mode"] then return false end

    local urls = {
        "/LAPI/V1.0/Channels/" .. self.channel .. "/Alarm/MotionDetection/WeekPlan",
        "/LAPI/V1.0/Channels/" .. self.channel .. "/Alarm/AudioDetection/WeekPlan"
    }

    for _, url in ipairs(urls) do
        local plan = http_request(url)
        if plan then
            plan.Enabled = bool_to_number(settings["general/online_mode"].value)
            if not http_request(url, "PUT", json.write(plan)) then return false end
        else
            return false
        end
    end

    return true
end

function setup:get_streams()
    local video_capabilities = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Media/Video/Capabilities")
    if not video_capabilities then return end

    local video_mode = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Media/Video/Mode")
    if not video_mode then return end

    local streams = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Media/Video/Streams/DetailInfos")
    if not streams then return end

    for i = 1, streams.Num do
        local stream = streams.VideoStreamInfos[i]
        local stream_capability = video_capabilities.StreamCapabilityList[i]

        if stream.Enabled == 1 and stream_capability then
            local stream_prefix = "streams/" .. tostring(i - 1) .. "/"

            -- get codec
            settings[stream_prefix .. "codec"].value = settings.codecs[tostring(stream.VideoEncodeInfo.EncodeFormat)]
            local codec_range = {}
            for k = 1, video_capabilities.EncodeFormatNum do
                local codec = settings.codecs[tostring(video_capabilities.EncodeFormatList[k])]
                if codec and not codec:find("JP") then
                    codec_range[#codec_range + 1] = codec
                end
            end
            settings[stream_prefix .. "codec"].meta.range = codec_range

            -- get resolution
            settings[stream_prefix .. "resolution"].value = string.format("%dx%d", stream.VideoEncodeInfo.Resolution.Width, stream.VideoEncodeInfo.Resolution.Height)
            local res_range = {}
            for k = 1, stream_capability.ResolutionNum do
                local resolution = stream_capability.ResolutionCapabilityList[k]
                res_range[k] = string.format("%dx%d", resolution.Width, resolution.Height)
            end
            settings[stream_prefix .. "resolution"].meta.range = res_range

            -- get bitrate
            settings[stream_prefix .. "bitrate"].value = stream.VideoEncodeInfo.BitRate
            local rate_range = {}
            for k = 1, stream_capability.ResolutionNum do
                local resolution = stream_capability.ResolutionCapabilityList[k]
                if resolution.Width == stream.VideoEncodeInfo.Resolution.Width and resolution.Height == stream.VideoEncodeInfo.Resolution.Height then
                    for _, bitrate in ipairs(settings.bitrates) do
                        if bitrate < resolution.MaxBitRate and bitrate >= resolution.MinBitRate then
                            rate_range[#rate_range + 1] = bitrate
                        end
                    end
                end
            end
            if not is_in_table(stream.VideoEncodeInfo.BitRate, rate_range) then rate_range[#rate_range + 1] = stream.VideoEncodeInfo.BitRate end
            settings[stream_prefix .. "bitrate"].meta.range = rate_range

            settings[stream_prefix .. "variable_bitrate"].value = (stream.VideoEncodeInfo.BitRateType == 1) -- bool

            -- get fps
            settings[stream_prefix .. "fps"].value = stream.VideoEncodeInfo.FrameRate
            local fps_range = {}
            for k, v in ipairs(settings.fpses) do
                if v > video_mode.FrameRate then break end
                fps_range[k] = v
            end
            settings[stream_prefix .. "fps"].meta.range = fps_range

            settings[stream_prefix .. "i_frame_interval"].value = stream.VideoEncodeInfo.IFrameInterval
            settings[stream_prefix .. "i_frame_interval"].meta.range = {video_capabilities.MinIFrameInterval, video_capabilities.MaxIFrameInterval}
        else
            for _, name in ipairs(stream_params) do
                settings:remove("streams/" .. tostring(i - 1) .. "/" .. name)
            end
        end
    end
end

function setup:set_streams()
    local streams = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Media/Video/Streams/DetailInfos")
    if not streams then return false end

    for i = 1, streams.Num do
        local stream = streams.VideoStreamInfos[i]
        if stream.Enabled ~= 1 then break end
        local enc = stream.VideoEncodeInfo -- helper
        local stream_prefix = "streams/" .. tostring(i - 1) .. "/"

        local codec = settings[stream_prefix .. "codec"]
        if codec and codec.changed then
            for k, v in pairs(settings.codecs) do
                if v == codec.value then
                    enc.EncodeFormat = tonumber(k)
                    break
                end
            end
        end

        local resolution = settings[stream_prefix .. "resolution"]
        if resolution and resolution.changed then
            width, height = resolution.value:match("(%d+)x(%d+)")
            width, height = tonumber(width), tonumber(height)

            enc.Resolution.Width = width
            enc.Resolution.Height = height
        end

        local bitrate = settings[stream_prefix .. "bitrate"]
        if bitrate and bitrate.changed then enc.BitRate = bitrate.value end

        local is_vbr = settings[stream_prefix .. "variable_bitrate"]
        if is_vbr and is_vbr.changed then enc.BitRateType = (is_vbr.value) and 1 or 0 end

        local fps = settings[stream_prefix .. "fps"]
        if fps and fps.changed then enc.FrameRate = fps.value end

        local ifi = settings[stream_prefix .. "i_frame_interval"]
        if ifi and ifi.changed then enc.IFrameInterval = ifi.value end
    end

    return http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Media/Video/Streams/DetailInfos", "PUT", json.write(streams))
end

function setup:get_timezone()
    local systime = http_request("/LAPI/V1.0/System/Time")
    if not systime then return end

    settings["general/timezone"].value = systime.TimeZone:match("GMT(%S+)")
    settings["general/timezone"].meta.range = settings.timezones
end

function setup:set_timezone()
    local systime = http_request("/LAPI/V1.0/System/Time")
    if not systime then return false end

    local timezone = settings["general/timezone"]
    if not timezone or not timezone.value then return false end

    systime.TimeZone = "GMT" .. timezone.value
    return http_request("/LAPI/V1.0/System/Time", "PUT", json.write(systime))
end

function setup:get_lamp()
    local ctrl = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Image/LampCtrl")
    if not ctrl then
        settings:remove("night_vision/adaptive/enabled")
        return
    end

    settings["night_vision/adaptive/enabled"].value = (ctrl.Enabled == 1)
end

function setup:set_lamp()
    if not settings["night_vision/adaptive/enabled"] then return false end
    local ctrl = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Image/LampCtrl")
    if not ctrl then return false end

    -- TODO: ctrl.Mode = 0 -- Light: 0=Infrared, 1=Visible

    if settings["night_vision/adaptive/enabled"].value == true then ctrl.Enabled = 1
    else ctrl.Enabled = 0 end

    return http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Image/LampCtrl", "PUT", json.write(ctrl))
end

function setup:get_night_mode()
    local video = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Image/Advanced/Exposure")
    if not table.is_subtable_seq(video, "DayNight") then
        settings:remove("night_vision/mode")
        return
    end

    if video.DayNight.Mode == 2 then
        settings["night_vision/mode"].value = "on"
    elseif video.DayNight.Mode == 1 then
        settings["night_vision/mode"].value = "off"
    else
        settings["night_vision/mode"].value = "auto"
    end
end

function setup:set_night_mode()
    local video = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Image/Advanced/Exposure")
    if not table.is_subtable_seq(video, "DayNight") then return false end

    local mode = settings["night_vision/mode"]
    if mode.value == "on" then video.DayNight.Mode = 2
    elseif mode.value == "off" then video.DayNight.Mode = 1
    else video.DayNight.Mode = 0 end

    return http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Image/Advanced/Exposure", "PUT", json.write(video))
end

function setup:get_audio_input()
    local input = http_request("/LAPI/V1.0/Media/Audio/Input")
    local caps = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Media/Audio/Capabilities")
    if not input or not caps or (caps.AudioInNum == 0) then
        settings:remove("microphone/enabled")
        settings:remove("microphone/sensitivity")
        settings:remove("microphone/denoise")
        settings:remove("streams/0/acodec")
        settings:remove("microphone/source")
        return
    end

    settings["microphone/enabled"].value = (input.IsMute ~= 1)

    if input.InputGain then
        settings["microphone/sensitivity"].value = input.InputGain
    else settings:remove("microphone/sensitivity") end

    if type(input.NoiseReduction) == "table" then
        settings["microphone/denoise"].value = (input.NoiseReduction.Enabled == 1)
    else settings:remove("microphone/denoise") end

    -- Set only 1st stream as primary mandatory stream:
    local ac = settings["streams/0/acodec"]
    if ac and caps.AudioInEncodeFormatNum > 1 then
        ac.value = settings.acodecs[input.EncodeFormat]
        if not ac.meta.range then
            local r = {}
            for _, v in ipairs(caps.AudioInEncodeFormatInfoList) do r[#r + 1] = settings.acodecs[v.Type] end
            ac.meta.range = r -- TODO: each codec has few fequencies from settings.afreqs
        end
    else
        settings:remove("streams/0/acodec")
    end

    local ms = settings["microphone/source"]
    if ms and input.AudioInputNum > 1 then -- instead of caps.AudioInNum
        local r = {}
        -- for _,v in ipairs(caps.AudioInModeInfoList) do for _, src in ipairs(v.ModeList) do r[#r + 1] = settings.asources[src] end end
        for _,v in ipairs(input.AudioInputList) do
            local src = settings.asources[v.Mode] -- support only DIFFERENT TYPES selection
            -- API by design allows to set multiple sources with the same type (few Mics, LineIns and etc.)
            if not is_in_table(src, r) then r[#r + 1] = settings.asources[v.Mode] end
            if (v.Enabled == 1) then ms.value = settings.asources[v.Mode] end
        end

        if #r < 2 then settings:remove("microphone/source")
        else ms.meta.range = r end
    else
        settings:remove("microphone/source")
    end
end

function setup:set_audio_input()
    local input = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Media/Audio/Input")
    if not input then return end

    if settings["microphone/enabled"] then input.IsMute = bool_to_number(not settings["microphone/enabled"].value) end
    if settings["microphone/sensitivity"] and settings["microphone/sensitivity"].changed then input.InputGain = settings["microphone/sensitivity"].value end
    if settings["microphone/denoise"] and settings["microphone/denoise"].changed then input.NoiseReduction.Enabled = (settings["microphone/denoise"].value) and 1 or 0 end

    if settings["streams/0/acodec"] and settings["streams/0/acodec"].changed then
        for k,v in pairs(settings.acodecs) do
            if v == settings["streams/0/acodec"].value then input.EncodeFormat = k ; break end
        end
    end

    if settings["microphone/source"] and settings["microphone/source"].changed then
        for _,v in ipairs(input.AudioInputList) do
            if settings["microphone/source"].value == settings.asources[v.Mode] then
                v.Enabled = 1
            else
                v.Enabled = 0
            end
        end
    end

    return http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Media/Audio/Input", "PUT", json.write(input))
end

function setup:get_motion_detector()
    local capabilities = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Alarm/Capabilities")
    if not capabilities then return end

    local area = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Alarm/MotionDetection/AreaType")
    if not area then return end

    if area.Type ~= 1 then
        area.Type = 1
        if not http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Alarm/MotionDetection/AreaType", "PUT", json.write(ctrl)) then
            settings:remove("motion_detector/enabled")
            return
        end
    end

    area = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Alarm/MotionDetection/Areas/Grid")
    if not area then return end

    settings["motion_detector/enabled"].value = (area.Enabled == 1)
    settings["motion_detector/0/sensitivity"].value = area.Sensitivity

    local data = base64.decode(area.Grid, true)
    local unpacked = { }

    local i = 1
    while i <= #data do
        local n = unsigned(data[i])
        if n > 128 then
            n = 1 - (n - 256)
            i = i + 1

            for j = 1, n do
                unpacked[#unpacked + 1] = unsigned(data[i])
            end
        elseif n < 128 then
            n = 1 + n
            for j = i + 1, i + n do
                unpacked[#unpacked + 1] = unsigned(data[j])
            end

            i = i + n
        end

        i = i + 1
    end

    local grid = ""
    for _, n in ipairs(unpacked) do
        grid = grid .. number_to_binary_string(n, 8)
    end

    local width = capabilities.MotionDetection.BlockWidth
    local height = capabilities.MotionDetection.BlockHeight

    settings["motion_detector/0/zones"].value = string.sub(grid, 0, width * height)
    settings["motion_detector/0/zones"].meta.range = { width, height }
end

function setup:set_motion_detector()
    local area = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Alarm/MotionDetection/Areas/Grid")
    if not area then return false end

    if settings["motion_detector/enabled"] then
        area.Enabled = bool_to_number(settings["motion_detector/enabled"].value)
    end

    local sensitivity = settings["motion_detector/0/sensitivity"]
    if sensitivity and sensitivity.value then
        area.Sensitivity = sensitivity.value
    end

    local zones = settings["motion_detector/0/zones"]
    if zones and zones.value then
        local value = zones.value .. string.rep("0", (zones.meta.range[1] * zones.meta.range[2]) % 8)

        local grid = { }
        local data = { }

        for bits in value:gmatch("........?") do
            grid[#grid + 1] = binary_string_to_number(bits)
        end

        local group = { }
        local prev = nil

        -- group bytes of data
        for _, value in pairs(grid) do
            if prev ~= value then
                group[#group + 1] = { }
            end

            table.insert(group[#group], value)
            prev = value
        end

        prev = { }
        for i = 1, #group + 1 do
            if group[i] == nil or #group[i] > 2 then
                if next(prev) ~= nil then
                    data[#data + 1] = #prev - 1
                    for _, value in pairs(prev) do
                        data[#data + 1] = value
                    end

                    prev = { }
                end

                if group[i] then
                    data[#data + 1] = 256 + (1 - #group[i])
                    data[#data + 1] = group[i][1]
                end
            else
                for _, value in pairs(group[i]) do
                    prev[#prev + 1] = value
                end
            end
        end

        area.Grid = base64.encode(data)
    end

    return http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Alarm/MotionDetection/Areas/Grid", "PUT", json.write(area))
end

function setup:get_audio_detector()
    local capabilities = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Alarm/Capabilities")
    if not capabilities or capabilities.AudioDetection.SupportCfg == 0 then
        settings:remove("sound_detection/enabled")
        settings:remove("sound_detection/sensitivity")
        return
    end

    local rule = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Alarm/AudioDetection/Rule")
    if not rule then return end

    local threshold = math.min(rule.DiffValue, rule.Threshold)
    threshold = tonumber(string.format("%.0f", threshold))

    settings["sound_detection/enabled"].value = (rule.Enabled == 1)
    settings["sound_detection/sensitivity"].value = round( (400 - threshold) / 4 )
    settings["sound_detection/sensitivity"].meta.range = {0, 100, 1}
end

function setup:set_audio_detector()
    local rule = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Alarm/AudioDetection/Rule")
    if not rule then return false end

    if settings["sound_detection/enabled"] then
        rule.Enabled = bool_to_number(settings["sound_detection/enabled"].value)
    end

    if settings["sound_detection/sensitivity"] then
        local threshold = (400 - 4 * settings["sound_detection/sensitivity"].value)
        rule.DiffValue = threshold
        rule.Threshold = threshold
    end

    return http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Alarm/AudioDetection/Rule", "PUT", json.write(rule))
end

function setup:get_osd()
    local capabilities = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Media/OSDs/Capabilities")
    if not capabilities then return end

    local num = capabilities.MaxAreaNum
    if num == 0 then
        settings:remove("osd/transparent")

        settings:remove("osd/datetime/enabled")
        settings:remove("osd/datetime/format")
        settings:remove("osd/datetime/pos")

        settings:remove("osd/name/enabled")
        settings:remove("osd/name/text")
        settings:remove("osd/name/pos")

        for i = 0, 5 do
            settings:remove("osd/custom/" .. tostring(i) .. "/text")
            settings:remove("osd/custom/" .. tostring(i) .. "/pos")
        end

        return
    end

    local style = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Media/OSDs/ContentStyle")
    if not style then return end

    local contents = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Media/OSDs/Contents")
    if not contents then return end

    local device_info = http_request("/LAPI/V1.0/System/DeviceInfo")
    if not device_info then return end

    if style.FontStyle then
        settings["osd/transparent"].value = (style.FontStyle == 3)
    else
        settings:remove("osd/transparent")
    end

    -- date&time
    local content = num >= 1 and contents.ContentList[device_info.DeviceType == 0 and 1 or 2] or nil
    local info = content and content.ContentInfo[1] or nil

    if content and content.Enabled == 1 and info.ContentType == 2 then
        settings["osd/datetime/enabled"].value = true
        settings["osd/datetime/pos"].meta.range = { 9999, 9999 }
        settings["osd/datetime/pos"].value = { content.Area.TopLeft.X, 9999 - content.Area.TopLeft.Y }

        local range = { }
        for i = 1, capabilities.SupportedDateFormatNum do
            for j = 1, capabilities.SupportedTimeFormatNum do
                local df = settings.date_format[capabilities.SupportedDateFormatList[i]]
                local tf = settings.time_format[capabilities.SupportedTimeFormatList[j]]

                if df and tf then
                    range[#range + 1] = df .. " " .. tf
                end
            end
        end

        local di = style.DateFormat
        local ti = style.TimeFormat

        settings["osd/datetime/format"].value = settings.date_format[di] .. " " .. settings.time_format[ti]
        settings["osd/datetime/format"].meta.range = range
    else
        settings["osd/datetime/enabled"].value = false
        settings:remove("osd/datetime/format")
        settings:remove("osd/datetime/pos")
    end

    -- ipc name
    content = num >= 2 and contents.ContentList[device_info.DeviceType == 0 and 2 or 1] or nil
    info = content and content.ContentInfo[1] or nil

    if content and content.Enabled == 1 and (info.ContentType == 1 or info.ContentType == 12) then
        settings["osd/name/enabled"].value = true
        settings["osd/name/text"].value = info.Value
        settings["osd/name/pos"].meta.range = { 9999, 9999 }
        settings["osd/name/pos"].value = { content.Area.TopLeft.X, 9999 - content.Area.TopLeft.Y }
    else
        settings["osd/name/enabled"].value = false
        settings:remove("osd/name/text")
        settings:remove("osd/name/pos")
    end

    -- custom text
    for i = 0, 5 do
        if i + 3 <= num then
            content = contents.ContentList[i + 3]
            info = content.ContentInfo[1]

            if content.Enabled == 1 and info.ContentType == 1 then
                settings["osd/custom/" .. tostring(i) .. "/text"].value = info.Value
            else
                settings["osd/custom/" .. tostring(i) .. "/text"].value = ""
            end

            settings["osd/custom/" .. tostring(i) .. "/pos"].meta.range = { 9999, 9999 }
            settings["osd/custom/" .. tostring(i) .. "/pos"].value = { content.Area.TopLeft.X, 9999 - content.Area.TopLeft.Y }
        else
            settings:remove("osd/custom/" .. tostring(i) .. "/text")
            settings:remove("osd/custom/" .. tostring(i) .. "/pos")
        end
    end
end

function setup:set_osd()
    local style = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Media/OSDs/ContentStyle")
    if not style then return false end

    local contents = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Media/OSDs/Contents")
    if not contents then return false end

    local device_info = http_request("/LAPI/V1.0/System/DeviceInfo")
    if not device_info then return end

    if settings["osd/transparent"] then
        if settings["osd/transparent"].value then
            style.FontStyle = 3
        else
            style.FontStyle = 0
        end
    end

    local dt_index = device_info.DeviceType == 0 and 1 or 2
    local nm_index = device_info.DeviceType == 0 and 2 or 1

    -- data&time
    if settings["osd/datetime/enabled"] then
        local enabled = settings["osd/datetime/enabled"].value
        contents.ContentList[dt_index].Enabled = bool_to_number(enabled)

        if enabled then
            contents.ContentList[dt_index].ContentInfo[1].ContentType = 2
        end
    end

    if settings["osd/datetime/pos"] then
        contents.ContentList[dt_index].Area.TopLeft.X = settings["osd/datetime/pos"].value[1]
        contents.ContentList[dt_index].Area.TopLeft.Y = 9999 - settings["osd/datetime/pos"].value[2]
    end

    if settings["osd/datetime/format"] then
        local format = settings["osd/datetime/format"].value

        local index
        for _, msg in pairs(settings.time_format) do
            index = string.gmatch(format, "()(" .. msg .. ")")()
            if index then break end
        end

        if not index then
            return false
        end

        local dt = string.sub(format, 0, index - 2)
        for key, value in pairs(settings.date_format) do
            if value == dt then
                style.DateFormat = key
                break
            end
        end

        local tm = string.sub(format, index)
        for key, value in pairs(settings.time_format) do
            if value == tm then
                style.TimeFormat = key
                break
            end
        end
    end

    -- ipc name
    if settings["osd/name/enabled"] then
        local enabled = settings["osd/name/enabled"].value
        contents.ContentList[nm_index].Enabled = bool_to_number(enabled)

        if enabled then
            -- content types:
            --     1 - custom
            --    12 - channel name
            contents.ContentList[nm_index].ContentInfo[1].ContentType = (device_info.DeviceType == 0 and 1 or 12)
        end
    end

    if settings["osd/name/text"] then
        contents.ContentList[nm_index].ContentInfo[1].Value = settings["osd/name/text"].value
    end

    if settings["osd/name/pos"] then
        contents.ContentList[nm_index].Area.TopLeft.X = settings["osd/name/pos"].value[1]
        contents.ContentList[nm_index].Area.TopLeft.Y = 9999 - settings["osd/name/pos"].value[2]
    end

    -- custom text
    for i = 0, 5 do
        if i + 3 > contents.Num then break end

        local content = contents.ContentList[i + 3]
        if settings["osd/custom/" .. tostring(i) .. "/text"] then
            local text = settings["osd/custom/" .. tostring(i) .. "/text"].value
            content.Enabled = bool_to_number(not string.is_empty(text))

            if content.Enabled == 1 then
                content.ContentInfo[1].Value = text
            end
        end

        if settings["osd/custom/" .. tostring(i) .. "/pos"] then
            content.Area.TopLeft.X = settings["osd/custom/" .. tostring(i) .. "/pos"].value[1]
            content.Area.TopLeft.Y = 9999 - settings["osd/custom/" .. tostring(i) .. "/pos"].value[2]
        end
    end

    if not http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Media/OSDs/ContentStyle", "PUT", json.write(style)) then
        return false
    end

    if not http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Media/OSDs/Contents", "PUT", json.write(contents)) then
        return false
    end

    return true
end

function setup:get_image_settings()
    local enhance = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Image/Enhance")
    if not enhance then return end

    settings["image/brightness"].value = enhance.Brightness
    settings["image/contrast"].value = enhance.Contrast
    settings["image/saturation"].value = enhance.Saturation

    -- Image rotation mode list:
    -- 0: Normal
    -- 1: Flip Vertical
    -- 2: Flip Horizontal
    -- 3: 180°
    -- 4: 90° Clockwise
    -- 5: 90° Anti-clockwise
    local caps = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Image/Capabilities")
    local rotation_modes = {}
    if table.is_subtable_seq(caps, "Enhance", "ImageRotationModeList") then
        rotation_modes = caps.Enhance.ImageRotationModeList
    end

    if is_in_table(3, rotation_modes) then settings["image/upside_down"].value = (enhance.ImageRotation == 3)
    else settings:remove("image/upside_down") end

    if is_in_table(4, rotation_modes) then settings["image/rotate"].value = (enhance.ImageRotation > 3)
    else settings:remove("image/rotate") end
end

function setup:set_image_settings()
    local enhance = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Image/Enhance")
    if not enhance then return false end

    if settings["image/brightness"] then
        enhance.Brightness = settings["image/brightness"].value
    end

    if settings["image/contrast"] then
        enhance.Contrast = settings["image/contrast"].value
    end

    if settings["image/saturation"] then
        enhance.Saturation = settings["image/saturation"].value
    end

    local setting = settings["image/upside_down"]
    enhance.ImageRotation = (setting and setting.value) and 3 or 0

    local caps = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Image/Capabilities")
    local anti_clockwise_supported = false
    if table.is_subtable_seq(caps, "Enhance", "ImageRotationModeList") then anti_clockwise_supported = is_in_table(5, caps.Enhance.ImageRotationModeList) end
    setting = settings["image/rotate"]
    if setting and setting.value then
        if (enhance.ImageRotation == 3) and anti_clockwise_supported then
            enhance.ImageRotation = 5
        else
            enhance.ImageRotation = 4 -- could reset upside_down
        end
    end

    return http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Image/Enhance", "PUT", json.write(enhance))
end

function setup:get_exposure()
    local exposure = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Image/Advanced/Exposure")
    local caps = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Image/Capabilities")

    if caps then
        caps = caps.Exposure
    end

    if not exposure or not caps then
        settings:remove("image/wdr")
        settings:remove("image/gain")
        settings:remove("image/shutter_time")
        return
    end

    if type(exposure.WideDynamic) == "table" and caps.WideDynamic.SupportWideDynamicCfg and settings["image/wdr"] then
        settings["image/wdr"].value = (exposure.WideDynamic.Mode == 1)
    else
        settings:remove("image/wdr")
    end

    if type(exposure.GainInfo) == "table" and settings["image/gain"] then
        settings["image/gain"].value = exposure.GainInfo.Gain
        settings["image/gain"].meta.range = {caps.MinGain, caps.MaxGain}
    else
        settings:remove("image/gain")
    end

    local st = settings["image/shutter_time"] -- helper
    if (caps.Shutter.SupportShutterTimeCfg ~= 1) or not st then
        settings:remove("image/shutter_time")
        return
    end

    st.value = "1/" .. tostring(exposure.ShutterInfo.Shutter)
    if exposure.Mode == 0 then st.value = "auto"
    elseif exposure.Mode == 4 then st.value = "gain"
    elseif exposure.Mode == 5 then st.value = "50hz"
    elseif exposure.Mode == 6 then st.value = "60hz"
    end
    -- 0: Auto 1: Custom 4: Gain priority 5: Indoor 50Hz 6: Indoor 60Hz 2: Shutter priority 7: Manual
    if not st.meta.range then
        local r = {}
        --local use_manual = false
        for _, v in ipairs(caps.ExposureModeList) do -- special modes
            if v == 0 then r[#r + 1] = "auto"
            elseif v == 4 then r[#r + 1] = "gain"
            elseif v == 5 then r[#r + 1] = "50hz"
            elseif v == 6 then r[#r + 1] = "60hz"
            --elseif v == 7 then use_manual = true
            end
        end
        --if use_manual then
        if (caps.Shutter.ShutterTimeUnit == 1) then -- else pure mks in range
            for _, v in ipairs(caps.Shutter.ShutterTimeRangeList) do r[#r + 1] = "1/" .. tostring(v) end
        else
            for _, v in ipairs(caps.Shutter.ShutterTimeRangeList) do r[#r + 1] = "1/" .. tostring(1000000/v) end
        end
        st.meta.range = r
        --end
    end
end

function setup:set_exposure()
    local exposure = http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Image/Advanced/Exposure")
    if not exposure then return false end

    if settings["image/wdr"] and settings["image/wdr"].changed then
        exposure.WideDynamic.Mode = (settings["image/wdr"].value) and 1 or 0
    end

    if settings["image/gain"] and settings["image/gain"].changed then
        exposure.GainInfo.Gain = settings["image/gain"].value
    end

    local st = settings["image/shutter_time"] -- helper
    if st and st.changed then
        if st.value:sub(1,2) == "1/" then
            exposure.Mode = 7
            exposure.ShutterInfo.Shutter = tonumber( st.value:sub(3) ) -- cut 1/ prefix
        elseif st.value == "gain" then exposure.Mode = 4
        elseif st.value == "50hz" then exposure.Mode = 5
        elseif st.value == "60hz" then exposure.Mode = 6
        else exposure.Mode = 0 end -- st.value == "auto"
    end

    return http_request("/LAPI/V1.0/Channels/" .. self.channel .. "/Image/Advanced/Exposure", "PUT", json.write(exposure))
end

--
-- INIT
--

function setup_api_request_params(cdi_url)
    http_request.base_url = (cdi_url or os.getenv("BASE_URL"))
    http_request.timeout_ms = 12000
    http_request.max_retries_count = 3
    http_request.retry_delay_ms = 1000
end

function setup_settings_handlers()
    table.merge(settings.traits, {
        ["general/online_mode"] = {
            get = setup.get_online_mode,
            set = setup.set_online_mode
        },
        ["general/timezone"] = {
            get = setup.get_timezone,
            set = setup.set_timezone
        },
        ["night_vision/mode"] = {
            get = setup.get_night_mode,
            set = setup.set_night_mode
        },
        ["night_vision/adaptive/enabled"] = {
            get = setup.get_lamp,
            set = setup.set_lamp
        },
        ["streams/0/acodec"] = {
            get = setup.get_audio_input,
            set = setup.set_audio_input
        },
        ["motion_detector/enabled"] = {
            get = setup.get_motion_detector,
            set = setup.set_motion_detector
        },
        ["motion_detector/0/sensitivity"] = {
            get = setup.get_motion_detector,
            set = setup.set_motion_detector
        },
        ["motion_detector/0/zones"] = {
            get = setup.get_motion_detector,
            set = setup.set_motion_detector
        },
        ["sound_detection/enabled"] = {
            get = setup.get_audio_detector,
            set = setup.set_audio_detector
        },
        ["sound_detection/sensitivity"] = {
            get = setup.get_audio_detector,
            set = setup.set_audio_detector
        }
    })

    for _, name in ipairs({"enabled", "sensitivity", "denoise", "source"}) do
        settings("microphone/" .. name).get = setup.get_audio_input
        settings("microphone/" .. name).set = setup.set_audio_input
    end

    for _, name in ipairs({ "transparent", "datetime/enabled", "datetime/format", "datetime/pos", "name/enabled", "name/text", "name/pos" }) do
        settings("osd/" .. name).get = setup.get_osd
        settings("osd/" .. name).set = setup.set_osd
    end

    for i = 0, 5 do
        settings("osd/custom/" .. tostring(i) .. "/text").get = setup.get_osd
        settings("osd/custom/" .. tostring(i) .. "/text").set = setup.set_osd
        settings("osd/custom/" .. tostring(i) .. "/pos").get = setup.get_osd
        settings("osd/custom/" .. tostring(i) .. "/pos").set = setup.set_osd
    end

    for i = 0, 2 do
        for _, name in ipairs(stream_params) do
            settings("streams/" .. tostring(i) .. "/" .. name).get = setup.get_streams
            settings("streams/" .. tostring(i) .. "/" .. name).set = setup.set_streams
        end
    end

    for _, name in ipairs({"brightness", "contrast", "saturation", "upside_down", "rotate"}) do
        settings("image/" .. name).get = setup.get_image_settings
        settings("image/" .. name).set = setup.set_image_settings
    end

    for _, name in ipairs({"wdr", "shutter_time", "gain"}) do
        settings("image/" .. name).get = setup.get_exposure
        settings("image/" .. name).set = setup.set_exposure
    end

    settings.traits.access_funcs_owner_table = setup
end

function get_info()
    local info = json.object()
    info.vendor = settings.vendor_name
    info.serial_number_format = "uniview"

    local device_info = http_request("/LAPI/V1.0/System/DeviceInfo")
    if device_info then
        info.serial_number = device_info.SerialNumber
        info.model = device_info.DeviceModel
        info.fw_version = device_info.FirmwareVersion
        info.hardware_id = device_info.HardwareID
    end

    local nework_interface = http_request("/LAPI/V1.0/Network/Interfaces/1")
    if nework_interface then
        info.mac_address = nework_interface.MAC
    end

    return info
end

function probe(cmd)
    http_request.max_retries_count = 1

    local reply = cmd
    if not cmd.args.inspection_protocol then
        cmd.args.inspection_protocol = cmd.args.cdi_scheme or "lapi"
    end

    local device_info, status = http_request("/LAPI/V1.0/System/DeviceInfo")
    if status ~= nil and status ~= 200 then
        reply.args = { status = status,  reason = "failed to request device info" }
        return reply
    end

    local nework_interface, status = http_request("/LAPI/V1.0/Network/Interfaces/1")
    if status ~= nil and status ~= 200 then
        reply.args = { status = status,  reason = "failed to request device network interfaces" }
        return reply
    end

    local result = cmd.args
    result.manufacturer = "Uniview"
    result.model = device_info.DeviceModel
    result.description = settings.device_type[device_info.DeviceType]
    result.presentation_url = (cmd.args.use_https and "https://" or "http://") .. cmd.args.host .. ":" .. tostring(cmd.args.port)
    result.sn = device_info.SerialNumber
    result.mac = nework_interface.MAC
    result.channels = { }

    local num_of_channels = 1
    -- todo: if device_info.DeviceType != 0 then get number of channels

    for i = 0, num_of_channels - 1 do
        local detail_infos, status = http_request("/LAPI/V1.0/Channels/" .. tostring(i) .. "/Media/Video/Streams/DetailInfos")
        if status ~= nil and status ~= 200 then
            reply.args = { status = status,  reason = "failed to request device video streams info" }
            return reply
        end

        local channel = { }
        channel.hw_id = tostring(i)
        channel.video = { }

        for n = 1, detail_infos.Num do
            if detail_infos.VideoStreamInfos[n].Enabled == 0 then
                goto continue
            end

            local id = detail_infos.VideoStreamInfos[n].ID
            local live_stream, status = http_request("/LAPI/V1.0/Channels/" .. tostring(i) .. "/Media/Video/Streams/" .. tostring(id) .. "/LiveStreamURL")
            if status ~= nil and status ~= 200 then
                reply.args = { status = status,  reason = "failed to request live stream url" }
                return reply
            end

            local stream_url = iv.copy_url_with(live_stream.URL, {userinfo = iv.build_url_userinfo(cmd.args.user, cmd.args.password)})
            if id == 0 then
                channel.video.primary = stream_url
            elseif id == 1 then
                channel.video.secondary = stream_url
            elseif id == 2 then
                channel.video.third = stream_url
            end

            ::continue::
        end

        channel.events = "onvif?source=video_source&port=" .. tostring(cmd.args.port)
        channel.config = cmd.args.cdi_scheme .. "?port=" .. tostring(cmd.args.port)

        local sd, status = http_request("/LAPI/V1.0/Channels/" .. tostring(i) .. "/System/DeviceStatus/SD")
        if status ~= nil and status ~= 200 and status ~= 400 then
            reply.args = { status = status,  reason = "failed to request device status of sd card" }
            return reply
        end

        if sd ~= nil and sd.StatusParam.SDStatus ~= 0 then
            channel.storage = http_request:copy_url_with{scheme = "uniview"}
        end

        local ptz, status = http_request("/LAPI/V1.0/Channels/" .. tostring(i) .. "/PTZ/Capabilities")
        if status ~= nil and status ~= 200 then
            reply.args = { status = status,  reason = "failed to request device ptz capabilities" }
            return reply
        end

        if ptz.IsSupportPTZ == 1 then
            channel.ptz = http_request:copy_url_with{scheme = "onvif-ptz"}
        end

        result.channels[#result.channels + 1] = channel
    end

    reply.args = { status = 200,  result = result}
    return reply
end

caps.requests.get_info = {
    useCache = true,
    level = "mod"
}
